import logging
from dataclasses import InitVar, field
from functools import cached_property
from pathlib import Path
from typing import Dict
from typing import Optional, Any, Union, Literal, List, Mapping, OrderedDict, Iterator

from dotenv import find_dotenv
from httpx import Client, BaseTransport, URL, Response
from httpx._client import EventHook
from httpx._types import CertTypes, VerifyTypes
from pydantic import ConfigDict, Field, InstanceOf, FilePath
from pydantic.dataclasses import dataclass
from pydantic_settings.sources import DotenvType

from ipfabric.auth import Setup, ProxyTypes, TimeoutTypes, ProxiesTypes
from ipfabric.exceptions import api_insuf_rights
from ipfabric.models import Snapshot, OAS, Endpoint, Snapshots, Methods, User
from ipfabric.tools import VALID_REFS, trigger_backup, raise_for_status

logger = logging.getLogger("ipfabric")

LAST_ID, PREV_ID, LASTLOCKED_ID = VALID_REFS


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class IPFabricAPI:
    """Initializes the IP Fabric Client.

    Args:
        base_url: IP Fabric instance provided in 'base_url' parameter, or the 'IPF_URL' environment variable
        api_version: [Optional] Version of IP Fabric API
        snapshot_id: IP Fabric snapshot ID to use by default for database actions - defaults to '$last'
        auth: API token, tuple (username, password), or custom Auth to pass to httpx
        unloaded: True to load metadata from unloaded snapshots
        env_file: Path to .env file to load
        streaming: Default True to use streaming instead of paging.
        verify: httpx.Client - Default True to verify IPF SSL certificate.
        timeout: httpx.Client - Default 5
        proxy: httpx.Client
        proxies: httpx.Client
        mounts: httpx.Client
        cert: httpx.Client
        event_hooks: httpx.Client
        local_oas: Default True, False mainly for development purposes
        local_oas_file: Default True, False mainly for development purposes
        debug: Future use.
    """

    base_url: Optional[Union[str, InstanceOf[URL]]] = Field(
        None, description="Base URL of the IPF instance", examples=["https://demo.ipfabric.io"]
    )
    api_version: Optional[str] = Field(None, description="Defaults to SDK or API version.")
    snapshot_id: Optional[str] = Field(None, description="Defaults to '$last'.")
    auth: InitVar[Optional[Any]] = field(default=None)
    unloaded: bool = Field(False, description="Default False, do not load unloaded snapshot metadata.")
    env_file: InitVar[Optional[DotenvType]] = field(default=None)
    streaming: bool = Field(True, description="Default True; use streaming results instead of pagination.")
    verify: InitVar[Optional[VerifyTypes]] = field(default=None)
    timeout: InitVar[Optional[Union[TimeoutTypes, Literal["DEFAULT"]]]] = field(default="DEFAULT")
    proxy: InitVar[Optional[ProxyTypes]] = field(default=None)
    proxies: InitVar[Optional[ProxiesTypes]] = field(default=None)  # TODO: 7.0 this is deprecated.
    mounts: InitVar[Optional[Mapping[str, Optional[BaseTransport]]]] = field(default=None)
    cert: InitVar[Optional[CertTypes]] = field(default=None)
    event_hooks: InitVar[Optional[Mapping[str, List[EventHook]]]] = field(default=None)
    local_oas: bool = Field(True, description="Default True, use local minified OAS file instead of servers.")
    local_oas_file: Optional[FilePath] = None
    debug: Optional[bool] = None
    http2: InitVar[Optional[bool]] = field(default=True)
    _os_version: Optional[str] = None
    _os_api_version: Optional[str] = None
    _client: Client = None
    _prev_snapshot_id: Optional[str] = None
    _attribute_filters: Optional[dict] = None
    _no_loaded_snapshots: bool = False
    _oas: Optional[OAS] = None
    _user: Optional[User] = None
    _snapshots: Optional[Snapshots] = None

    def __post_init__(
        self,
        auth,
        env_file,
        verify,
        timeout,
        proxy,
        proxies,
        mounts,
        cert,
        event_hooks,
        http2,
    ):
        env_file = env_file if env_file else find_dotenv(usecwd=True) or Path("~").expanduser().joinpath(".env")
        logger.info(f"Using .env file located at {env_file}.")
        setup = Setup(
            base_url=self.base_url,
            api_version=self.api_version,
            auth=auth,
            _env_file=env_file,
            snapshot_id=self.snapshot_id,
            verify=verify,
            timeout=timeout,
            proxy=proxy,
            proxies=proxies,
            mounts=mounts,
            cert=cert,
            event_hooks=event_hooks,
            debug=self.debug,
            http2=http2,
        )
        [setattr(self, k, v) for k, v in setup.update_attrs.items()]

        self._oas = OAS(client=self, local_oas=self.local_oas, local_oas_file=self.local_oas_file)
        self._user = self.get_user()
        self._snapshots = Snapshots(client=self)
        self.snapshot_id = setup.snapshot_id
        logger.debug(
            f"Successfully connected to '{self.base_url.host}' IPF version '{self.os_version}' "
            f"as user '{self.user.username}'"
        )

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "timeout":
            self._client.timeout = value
        elif name == "snapshot_id":
            value = self._switch_snapshot(value)
        super().__setattr__(name, value)

    def _switch_snapshot(self, snapshot_id: str) -> Union[str, None]:
        """Verify snapshot ID is valid."""
        if not self.loaded_snapshots:
            logger.warning("No Snapshots are currently loaded.  Please load a snapshot before querying any data.")
            self._no_loaded_snapshots = True
            return None
        elif snapshot_id in self.snapshots:
            self._prev_snapshot_id = self.snapshot_id
            return self.snapshots[snapshot_id].snapshot_id
        else:
            raise ValueError(f"Incorrect Snapshot ID: '{snapshot_id}'")

    def get(self, url, *args, params=None, **kwargs) -> Response:
        return self._client.get(url, *args, params=params, **kwargs)

    def post(self, url, *args, json=None, **kwargs) -> Response:
        return self._client.post(url, *args, json=json, **kwargs)

    def put(self, url, *args, json=None, **kwargs) -> Response:
        return self._client.put(url, *args, json=json, **kwargs)

    def patch(self, url, *args, json=None, **kwargs) -> Response:
        return self._client.patch(url, *args, json=json, **kwargs)

    def request(self, method, url, *args, json=None, **kwargs) -> Response:
        return self._client.request(method, url, *args, json=json, **kwargs)

    def delete(self, url, *args, **kwargs) -> Response:
        return self._client.delete(url, *args, **kwargs)

    def stream(self, method, url, *args, **kwargs) -> Iterator[Response]:
        return self._client.stream(method, url, *args, **kwargs)

    @property
    def os_version(self):
        return self._os_version

    @property
    def os_api_version(self):
        return self._os_api_version

    @property
    def user(self) -> User:
        return self._user

    @cached_property
    def _api_insuf_rights(self):
        return api_insuf_rights(self._user)

    @cached_property
    def hostname(self) -> str:
        resp = self.get("/os/hostname")
        if resp.status_code == 200:
            return resp.json()["hostname"]
        else:
            logger.critical(self._api_insuf_rights + 'on GET "/os/hostname"; Using URL host.')
            return str(self.base_url.host)

    @property
    def oas(self) -> Dict[str, Methods]:
        return self._oas.oas

    @property
    def web_to_api(self) -> Dict[str, Endpoint]:
        return self._oas.web_to_api

    @property
    def scope_to_api(self) -> Dict[str, Endpoint]:
        return self._oas.scope_to_api

    @property
    def attribute_filters(self):
        return self._attribute_filters

    @attribute_filters.setter
    def attribute_filters(self, attribute_filters: Union[Dict[str, List[str]], None]):
        if attribute_filters:
            logger.warning(
                "Setting Global Attribute Filter for all tables/diagrams until explicitly unset to None.\n"
                "Adding an Attribute Filter to any function will overwrite the Global Filter.\n"
                f"Filter: {attribute_filters}"
            )
        self._attribute_filters = attribute_filters

    @property
    def snapshots(self) -> OrderedDict[str, Snapshot]:
        return self._snapshots.snapshots

    @property
    def snapshot(self) -> Snapshot:
        return self.snapshots[self.snapshot_id]

    @property
    def loaded_snapshots(self) -> OrderedDict[str, Snapshot]:
        """get only loaded snapshots"""
        return self._snapshots.loaded_snapshots

    @property
    def unloaded_snapshots(self) -> OrderedDict[str, Snapshot]:
        return self._snapshots.unloaded_snapshots

    @property
    def loading_snapshot(self) -> Union[Snapshot, None]:
        """Return Loading Snapshot or None"""
        return self._snapshots.loading_snapshot

    @property
    def running_snapshot(self) -> Union[Snapshot, None]:
        """Return Running Snapshot"""
        return self._snapshots.running_snapshot

    def get_snapshot(self, snapshot_id: str) -> Snapshot:
        return self._snapshots.get_snapshot(snapshot_id)

    def get_snapshot_id(self, snapshot: Union[Snapshot, str]) -> str:
        return self._snapshots.get_snapshot_id(snapshot)

    def get_snapshots(self) -> OrderedDict[str, Snapshot]:
        return self._snapshots.get_snapshots()

    @property
    def prev_snapshot_id(self) -> str:
        """get previous snapshot Id"""
        return self._prev_snapshot_id

    def get_user(self) -> User:
        """Gets current logged in user information.

        Returns:
            User: User model of logged in user
        """
        resp = raise_for_status(self.get("users/me"))
        user = User(**resp.json())
        if not (user.is_admin and user.token):  # TODO: NIM-14008: Implement logic after obtaining scopes for tokens
            resp = raise_for_status(self.get("users/me/scopes/api"))
            user.scopes = resp.json()["data"]
        return user

    def _ipf_pager(
        self,
        url: str,
        payload: dict,
        limit: int = 1000,
        start: int = 0,
    ):
        """
        Loops through and collects all the data from the tables
        :param url: str: Full URL to post to
        :param payload: dict: Data to submit to IP Fabric
        :param start: int: Where to start for the data
        :return: list: List of dictionaries
        """
        payload["pagination"] = dict(limit=limit)
        data = list()

        def page(s):
            payload["pagination"]["start"] = s
            r = raise_for_status(self.post(url, json=payload))
            return r.json()["data"]

        r_data = page(start)
        data.extend(r_data)
        while limit == len(r_data):
            start = start + limit
            r_data = page(start)
            data.extend(r_data)
        return data

    def trigger_backup(self, sn: str = None, ip: str = None):
        return trigger_backup(self, sn=sn, ip=ip)



import ipaddress
import json
import logging
from typing import List, Any, Union, Dict, Literal
from collections import OrderedDict

from macaddress import EUI48
from pydantic import BaseModel, Field, PrivateAttr

from ipfabric.models.oas import Endpoint, NestedColumn
from ipfabric.tools.shared import parse_mac

LOGGER = logging.getLogger("ipfabric")

VM_INTS = "/tables/cloud/virtual-machines-interfaces"
RANKED = {
    "ipv4": [
        "/tables/addressing/managed-devs",
        "/tables/inventory/interfaces",
        "/tables/addressing/hosts",
        "/tables/addressing/arp",
        "/tables/wireless/clients",
        "/tables/inventory/phones",
        VM_INTS,
        "/tables/neighbors/unmanaged",
        "/tables/interfaces/connectivity-matrix/unmanaged-neighbors/detail",
        "/tables/routing/protocols/bgp/neighbors",
        "/tables/routing/protocols/eigrp/neighbors",
        "/tables/routing/protocols/is-is/neighbors",
        "/tables/routing/protocols/ospf/neighbors",
        "/tables/routing/protocols/rip/neighbors",
    ],
    "ipv6": [
        "/tables/addressing/ipv6-managed-devs",
        "/tables/addressing/ipv6-hosts",
        "/tables/addressing/ipv6-neighbors",
        "/tables/neighbors/unmanaged",
        "/tables/interfaces/connectivity-matrix/unmanaged-neighbors/detail",
        VM_INTS,
        "/tables/routing/protocols/bgp/neighbors",
        "/tables/routing/protocols/is-is/neighbors",
        "/tables/routing/protocols/ospf-v3/neighbors",  # NIM-15216
    ],
    "mac": [
        "/tables/addressing/managed-devs",
        "/tables/addressing/ipv6-managed-devs",
        "/tables/inventory/interfaces",
        "/tables/addressing/hosts",
        "/tables/addressing/ipv6-hosts",
        "/tables/addressing/ipv6-neighbors",
        "/tables/addressing/arp",
        VM_INTS,
        "/tables/addressing/mac",
        "/tables/wireless/radio",
    ],
}


class GlobalSearch(BaseModel):  # TODO: CIDR and Route Table Searching, Stop after first match
    client: Any = Field(exclude=True)
    _ipv4: List[Endpoint] = PrivateAttr(default_factory=list)
    _ipv6: List[Endpoint] = PrivateAttr(default_factory=list)
    _mac: List[Endpoint] = PrivateAttr(default_factory=list)

    def model_post_init(self, __context: Any) -> None:
        self._load_default()
        self._load_dynamic()

    def _load_default(self):
        for k, v in RANKED.items():
            for path in v:
                if getattr(self.client.oas[path[1:]].post, k).columns:
                    getattr(self.client.oas[path[1:]].post, k).full_scan = False
                    getattr(self, k).append(self.client.oas[path[1:]].post)

    def _load_dynamic(self):
        for path in self.client.oas.values():
            if path.post:
                for k, paths in RANKED.items():
                    if path.full_api_endpoint not in paths and getattr(path.post, k).columns:
                        getattr(self, k).append(path.post)

    @property
    def ipv4(self):
        return self._ipv4

    @property
    def ipv6(self):
        return self._ipv6

    @property
    def mac(self):
        return self._mac

    @staticmethod
    def _create_filter(
        columns: List[Union[str, NestedColumn]], arrays: List[str], value: str, regex: bool = False
    ) -> dict:
        oper = "reg" if regex else "eq"
        filters = []
        for col in columns:
            if isinstance(col, NestedColumn):
                filters.append({col.parent: ["any", col.child, oper, value]})
            elif col in arrays:
                filters.append({col: ["any", oper, value]})
            else:
                filters.append({col: [oper, value]})
        return {"or": filters}

    def _search(
        self,
        search_type: Literal["ipv4", "ipv6", "mac"],
        address: str,
        full_scan: bool = False,
        regex: bool = False,
    ) -> Dict[str, Dict[str, Union[str, list]]]:

        results = OrderedDict()

        for path in getattr(self, search_type):
            search_data = getattr(path, search_type)
            if not full_scan and search_data.full_scan is True:
                LOGGER.debug('Finished searching default tables; "--full-scan" not enabled to search all tables.')
                break  # List is ordered and once this is met we can break if not full scan
            filters = self._create_filter(search_data.columns, path.array_columns, address, True if regex else False)
            msg = f'earching "{path.web_menu}": API Endpoint: "{path.full_api_endpoint}"'
            LOGGER.debug(f'S{msg}; Filters `{json.dumps(filters, separators=(",", ":"))}`.')
            results[path.full_api_endpoint] = {
                "data": self.client.fetch_all(path.api_endpoint, filters=filters),
                "path": path.full_api_endpoint,
                "webPath": path.web_endpoint,
                "url": path.filter_url(filters),
                "menu": f'{path.web_menu}: "{path.title or path.summary}"',
            }
            LOGGER.debug(f"Finished s{msg}.")

        return results

    def search(
        self,
        address: str,
        full_scan: bool = False,
    ) -> Union[None, Dict[str, Dict[str, Union[str, list]]]]:

        if isinstance(address, int):
            raise TypeError(f"Input must be a valid string not integer: {str(address)}")
        try:
            return self.search_mac(address, full_scan)
        except ValueError:
            pass
        try:
            return self.search_ip(address, full_scan)
        except ValueError:
            raise SyntaxError(f'Address does not appear to be a IPv4, IPv6, nor MAC Address: "{address}".')

    def search_mac(
        self,
        address: str,
        full_scan: bool = False,
    ) -> Dict[str, Dict[str, Union[str, list]]]:

        if isinstance(address, int):
            raise TypeError(f"Input must be a valid string not integer: {str(address)}")
        LOGGER.info("Verifying Address is a MAC.")
        EUI48(address)
        mac = parse_mac(address)
        LOGGER.debug(f'Searching for MAC Address "{mac}".')
        return self._search("mac", mac, full_scan)

    def search_ip(
        self,
        address: str,
        full_scan: bool = False,
    ) -> Dict[str, Dict[str, Union[str, list]]]:
        LOGGER.info("Verifying Address is an IP.")
        ip = ipaddress.ip_address(address)
        if ip.version == 4:
            LOGGER.debug(f'Searching for IPv4 Address "{str(ip)}".')
            return self._search("ipv4", str(ip), full_scan)
        elif ip.version == 6:
            LOGGER.debug(f'Searching for IPv6 Address "{str(ip)}".')
            return self._search("ipv6", str(ip), full_scan)

    def _search_ip(
        self,
        version: int,
        address: str,
        full_scan: bool = False,
    ) -> Dict[str, Dict[str, Union[str, list]]]:
        ip = ipaddress.ip_address(address)
        if ip.version != version:
            raise ValueError()
        return self._search("ipv" + str(version), str(ip), full_scan)

    def search_ipv4(
        self,
        address: str,
        full_scan: bool = False,
    ) -> Union[None, Dict[str, Dict[str, Union[str, list]]]]:
        return self._search_ip(4, address, full_scan)

    def search_ipv6(
        self,
        address: str,
        full_scan: bool = False,
    ) -> Union[None, Dict[str, Dict[str, Union[str, list]]]]:
        return self._search_ip(6, address, full_scan)

    def search_regex(
        self,
        search_type: Literal["ipv4", "ipv6", "mac"],
        address: str,
        full_scan: bool = False,
    ) -> Dict[str, Dict[str, Union[str, list]]]:
        LOGGER.debug(f'Searching for {search_type.upper()} Address "{address}".')
        return self._search(search_type, address, full_scan, True)



from .device import Device, Devices
from .intent import Intent
from .inventory import Inventory
from .jobs import Jobs
from .oas import OAS, Endpoint, Methods
from .rbac import Role, Policy
from .snapshot import Snapshot, SNAPSHOT_COLUMNS
from .snapshots import Snapshots
from .table import BaseTable, Table
from .technology import Technology
from .users import User
from .global_search import GlobalSearch

__all__ = [
    "BaseTable",
    "Table",
    "Inventory",
    "Technology",
    "Jobs",
    "Snapshot",
    "Intent",
    "SNAPSHOT_COLUMNS",
    "Device",
    "Devices",
    "OAS",
    "Endpoint",
    "Methods",
    "Snapshots",
    "User",
    "Role",
    "Policy",
    "GlobalSearch",
]



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class Qos(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    @computed_field
    @property
    def policy_maps(self) -> Table:
        return Table(client=self.client, endpoint="tables/qos/policy-maps", sn=self.sn)

    @computed_field
    @property
    def shaping(self) -> Table:
        return Table(client=self.client, endpoint="tables/qos/shaping", sn=self.sn)

    @computed_field
    @property
    def queuing(self) -> Table:
        return Table(client=self.client, endpoint="tables/qos/queuing", sn=self.sn)

    @computed_field
    @property
    def policing(self) -> Table:
        return Table(client=self.client, endpoint="tables/qos/policing", sn=self.sn)

    @computed_field
    @property
    def priority_queuing(self) -> Table:
        return Table(client=self.client, endpoint="tables/qos/priority-queuing", sn=self.sn)

    @computed_field
    @property
    def marking(self) -> Table:
        return Table(client=self.client, endpoint="tables/qos/marking", sn=self.sn)

    @computed_field
    @property
    def random_drops(self) -> Table:
        return Table(client=self.client, endpoint="tables/qos/random-drops", sn=self.sn)



import logging
from typing import Optional, Dict, List, Any, Union, overload, Literal

from pydantic import BaseModel, computed_field, Field

from .device import Devices
from .table import Table

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None
logger = logging.getLogger("ipfabric")

IGNORE_COLUMNS = {"id"}
EXPORT_FORMAT = Literal["json", "csv", "df"]
DEVICE_EXPORT_FORMAT = Literal["json", "csv", "object", "df"]


class DeviceTable(Table):
    """model for Device Table data"""

    endpoint: str = "tables/inventory/devices"

    def _as_model(self, devices, export):
        if export != "object":
            return devices
        return Devices(
            client=self.client,
            snapshot_id=self.client.snapshot_id,
            devices=devices,
        )

    @overload
    def fetch(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> List[dict]: ...

    @overload
    def fetch(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> DataFrame: ...

    @overload
    def fetch(
        self,
        export: Literal["object"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> Devices: ...

    def fetch(
        self,
        export: DEVICE_EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv, object]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters'
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            limit: Default to 1,000 rows
            start: Starts at 0
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], Devices, bytes, DataFrame]: Default List of dicts 'json', Devices object if 'object',
                                                          bytes if 'csv', pandas.DataFrame if 'df'
        """
        devices = super(DeviceTable, self).fetch(
            export="json" if export in ["json", "object"] else export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports if export not in ["csv", "object"] else None,
            sort=sort,
            limit=limit,
            start=start,
            csv_tz=csv_tz,
        )
        return self._as_model(devices, export)

    @overload
    def all(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> List[dict]: ...

    @overload
    def all(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def all(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> DataFrame: ...

    @overload
    def all(
        self,
        export: Literal["object"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        client: Optional[Any] = None,
    ) -> Devices: ...

    def all(
        self,
        export: DEVICE_EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv, object]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
             Union[List[dict], Devices, bytes, DataFrame]: Default List of dicts 'json', Devices object if 'object',
                                                           bytes if 'csv', pandas.DataFrame if 'df'
        """
        devices = super(DeviceTable, self).all(
            export="json" if export in ["json", "object"] else export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports if export not in ["csv", "object"] else None,
            sort=sort,
            csv_tz=csv_tz,
        )
        return self._as_model(devices, export)


class Inventory(BaseModel):
    """model for inventories"""

    client: Any = Field(None, exclude=True)

    @computed_field
    @property
    def sites(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/sites")

    @computed_field
    @property
    def vendors(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/vendors")

    @computed_field
    @property
    def devices(self) -> Table:
        return DeviceTable(client=self.client)

    @computed_field
    @property
    def models(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/models")

    @computed_field
    @property
    def os_version_consistency(self) -> Table:
        return Table(client=self.client, endpoint="tables/management/osver-consistency")

    @computed_field
    @property
    def eol_summary(self) -> Table:
        return Table(client=self.client, endpoint="tables/reports/eof/summary")

    @computed_field
    @property
    def eol_details(self) -> Table:
        return Table(client=self.client, endpoint="tables/reports/eof/detail")

    @computed_field
    @property
    def platforms(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/platforms")

    @computed_field
    @property
    def pn(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/pn")

    @computed_field
    @property
    def families(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/summary/families")

    @computed_field
    @property
    def interfaces(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/interfaces")

    @computed_field
    @property
    def hosts(self) -> Table:
        return Table(client=self.client, endpoint="tables/addressing/hosts")

    @computed_field
    @property
    def hosts_ipv6(self) -> Table:
        return Table(client=self.client, endpoint="tables/addressing/ipv6-hosts")

    @computed_field
    @property
    def phones(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/phones")

    @computed_field
    @property
    def fans(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/fans")

    @computed_field
    @property
    def modules(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/modules")



import logging
import re
from http.cookiejar import CookieJar
from importlib import metadata
from typing import Optional, Union, Generator, Any, Type, Tuple, Literal, Mapping, List, Dict

from httpx import Client, BaseTransport, URL, Request, Response, Auth, Timeout, post, Proxy
from httpx._client import EventHook
from httpx._types import CertTypes, VerifyTypes, URLTypes
from pydantic import field_validator, Field, AliasChoices, PrivateAttr, InstanceOf
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource, InitSettingsSource

from ipfabric.tools import raise_for_status, valid_snapshot

logger = logging.getLogger("ipfabric")

RE_VERSION = re.compile(r"v?(\d(\.\d)?)")
ProxyTypes = Union[None, URLTypes, InstanceOf[Proxy]]
ProxiesTypes = Union[ProxyTypes, Dict[URLTypes, Union[None, ProxyTypes]]]
TimeoutTypes = Union[
    Optional[float], Tuple[Optional[float], Optional[float], Optional[float], Optional[float]], Timeout, None
]


def log_request(request: Request):
    logger.debug(f"Request event hook: {request.method} {request.url} - Waiting for response")


def log_response(response: Response):
    request = response.request
    logger.debug(f"Response event hook: {request.method} {request.url} - Status {response.status_code}")


class TokenAuth(Auth):
    def __init__(self, token: str) -> None:
        self._auth_header = token

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.headers["X-API-Token"] = self._auth_header
        yield request


class AccessToken(Auth):
    def __init__(self, cookie_jar: CookieJar, username: str, password: str, base_url: URL, verify: VerifyTypes):
        self.cookie_jar = cookie_jar
        self._login(username, password, base_url, verify)

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        response = yield request

        if response.status_code == 401:
            response.read()
            if "API_EXPIRED_ACCESS_TOKEN" in response.text:
                # Use refreshToken in Cookies to get new accessToken & Response updates accessToken in shared CookieJar
                resp = raise_for_status(post(response.url.join("/api/auth/token"), cookies=self.cookie_jar))
                request.headers["Cookie"] = "accessToken=" + resp.cookies["accessToken"]  # Update request
                yield request
        return response

    def _login(self, username: str, password: str, base_url: URL, verify: VerifyTypes) -> None:
        raise_for_status(
            post(
                base_url.join("auth/login"),
                json=dict(username=username, password=password),
                cookies=self.cookie_jar,
                verify=verify,
            )
        )


class MyInitSettingsSource(InitSettingsSource):
    def __init__(self, settings_cls, init_kwargs: Dict[str, Any]):
        timeout = init_kwargs.pop("timeout", "DEFAULT")
        init_kwargs = {k: v for k, v in init_kwargs.items() if v is not None}
        if timeout != "DEFAULT":
            init_kwargs["timeout"] = timeout
        super().__init__(settings_cls, init_kwargs)


class Setup(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", env_prefix="ipf_", extra="allow")
    base_url: Union[str, URL] = Field(None, validation_alias=AliasChoices("base_url", "ipf_url"))
    api_version: Optional[Union[int, float, str]] = Field(
        None, validation_alias=AliasChoices("api_version", "ipf_version")
    )
    auth: Optional[Any] = Field(None, alias="auth", exclude=True)
    token: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    snapshot_id: Optional[Union[str, None]] = Field(
        "$last", validation_alias=AliasChoices("snapshot_id", "ipf_snapshot")
    )
    verify: Optional[VerifyTypes] = True
    timeout: Union[TimeoutTypes, Literal["DEFAULT"]] = Timeout(timeout=5.0)
    nvd_api_key: Optional[str] = Field(None, alias="nvd_api_key")
    debug: Optional[bool] = False
    http2: Optional[bool] = True
    proxy: Optional[Union[ProxyTypes, None]] = None
    proxies: Optional[Union[ProxiesTypes, None]] = None
    mounts: Optional[Mapping[str, Optional[BaseTransport]]] = None
    cert: Optional[Union[CertTypes, None]] = None
    event_hooks: Optional[Mapping[str, List[EventHook]]] = None
    _cookie_jar: CookieJar = PrivateAttr(default_factory=CookieJar)
    client: Optional[Client] = Field(None, exclude=True)
    os_version: Optional[str] = Field(None)
    _os_api_version: Optional[str] = PrivateAttr(None)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (
            MyInitSettingsSource(settings_cls, init_settings.init_kwargs),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )

    def model_post_init(self, __context):
        event_hooks = {"request": [log_request], "response": [log_response]} if self.debug else self.event_hooks
        self.client = Client(
            base_url=self.base_url,
            cookies=self._cookie_jar,
            http2=self.http2,
            verify=self.verify,
            timeout=self.timeout,
            headers={"User-Agent": f'python-ipfabric-sdk/{metadata.version("ipfabric")}'},
            event_hooks=event_hooks,
            proxy=self.proxy,
            proxies=self.proxies,
            mounts=self.mounts,
            cert=self.cert,
        )
        if not (self.auth or self.token) and not (self.username and self.password):
            raise RuntimeError("IP Fabric Authentication not provided.")
        self.api_version, self.os_version, self._os_api_version = self.check_version()
        self.base_url = self.base_url.join(f"api/{self.api_version}/")
        self.client.base_url = self.base_url
        if self.auth:
            if isinstance(self.auth, str):
                self.client.auth = TokenAuth(self.auth)
            elif isinstance(self.auth, tuple):
                self.client.auth = AccessToken(self._cookie_jar, self.auth[0], self.auth[1], self.base_url, self.verify)
            else:
                self.client.auth = self.auth
        elif self.token:
            self.client.auth = TokenAuth(self.token)
        elif self.username and self.password:
            self.client.auth = AccessToken(self._cookie_jar, self.username, self.password, self.base_url, self.verify)

    @property
    def update_attrs(self) -> dict:
        return {
            "base_url": self.base_url,
            "api_version": self.api_version,
            "_os_version": self.os_version,
            "_os_api_version": self._os_api_version,
            "debug": self.debug,
            "_client": self.client,
            "timeout": self.client.timeout,
        }

    def check_version(self) -> tuple:
        """Checks API Version and returns the version to use in the URL and the OS Version

        Returns:
            api_version, os_version
        """
        cfg_version = self.api_version.lstrip("v").split(".")

        resp = raise_for_status(self.client.get("/api/version"))
        os_api_version = resp.json()["apiVersion"].lstrip("v").split(".")
        api_version = f"v{cfg_version[0]}.{cfg_version[1]}" if len(cfg_version) > 1 else f"v{cfg_version[0]}"
        if len(cfg_version) == 1 and cfg_version[0] > os_api_version[0]:
            logger.warning(
                f"Specified API or SDK Version (v{cfg_version[0]}) is greater then "
                f"OS API Version. Using OS Version:  (v{os_api_version[0]})"
            )
            api_version = f"v{os_api_version[0]}"
        elif cfg_version[0:2] > os_api_version:
            logger.warning(
                f"Specified API or SDK Version (v{'.'.join(cfg_version)}) is greater then "
                f"OS API Version. Using OS Version:  (v{'.'.join(os_api_version)})"
            )
            api_version = f"v{os_api_version[0]}.{os_api_version[1]}"
        elif os_api_version[0] > cfg_version[0]:
            raise RuntimeError(
                f"OS Major Version v{os_api_version[0]} is greater then SDK Version "
                f"v{cfg_version[0]}.  Please upgrade the Python SDK to the new major version."
            )

        return api_version, resp.json()["releaseVersion"], resp.json()["apiVersion"]

    @field_validator("api_version")
    @classmethod
    def _valid_version(cls, v: Union[None, int, float, str]) -> Union[None, str]:
        if not v:
            return metadata.version("ipfabric")
        elif RE_VERSION.match(str(v)):
            return "v" + RE_VERSION.match(str(v)).group(1)
        else:
            raise ValueError(f"IPF_VERSION ({v}) is not valid, must be like `v#` or `v#.#`.")

    @field_validator("base_url")
    @classmethod
    def _convert_url(cls, v: Union[URL, str]) -> URL:
        if isinstance(v, str):
            v = URL(v)
        return v

    @field_validator("snapshot_id")
    @classmethod
    def _valid_snapshot(cls, v: Union[None, str]) -> str:
        return valid_snapshot(v, init=True)

    @field_validator("verify")
    @classmethod
    def _verify(cls, v: Union[bool, int, str]) -> Union[bool, str]:
        if isinstance(v, bool):
            return v
        elif isinstance(v, int):
            return bool(v)
        elif v.lower() in {"0", "off", "f", "false", "n", "no", "1", "on", "t", "true", "y", "yes"}:
            return False if v.lower() in {0, "0", "off", "f", "false", "n", "no"} else True
        else:
            return v

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Needed for context"""
        pass



import logging
from typing import Any, Optional, Union

from httpx import HTTPStatusError
from pydantic import BaseModel, ConfigDict, Field

from .api_tokens import APIToken
from .attributes import Attributes
from .authentication import Authentication
from .discovery import Discovery
from .local_users import LocalUsers
from .rbac import Roles, Policies
from .seeds import Seeds
from .site_separation import SiteSeparation
from .vendor_api import VendorAPI

logger = logging.getLogger("ipfabric")


class Settings(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    client: Any = Field(exclude=True)
    _local_users: Optional[Union[LocalUsers, bool]] = None
    _api_tokens: Optional[Union[APIToken, bool]] = None
    _site_separation: Optional[Union[SiteSeparation, bool]] = None
    _discovery: Optional[Union[Discovery, bool]] = None
    _authentication: Optional[Union[Authentication, bool]] = None
    _seeds: Optional[Union[Seeds, bool]] = None
    _vendor_api: Optional[Union[VendorAPI, bool]] = None
    _global_attributes: Optional[Union[Attributes, bool]] = None
    _roles: Optional[Union[Roles, bool]] = None
    _policies: Optional[Union[Policies, bool]] = None

    def _try_object(self, obj, name, attr):
        if getattr(self, attr) is None:
            try:
                setattr(self, attr, obj(client=self.client))
            except HTTPStatusError:
                logger.warning(f"{self.client._api_insuf_rights} for `{name}` class.")
                setattr(self, attr, False)
        elif getattr(self, attr) is False:
            logger.warning(f"{self.client._api_insuf_rights} for `{name}` class.")
            return None
        return getattr(self, attr)

    @property
    def local_users(self) -> LocalUsers:
        return self._try_object(LocalUsers, 'LocalUsers', '_local_users')

    @property
    def roles(self) -> Roles:
        return self._try_object(Roles, 'Roles', '_roles')

    @property
    def policies(self) -> Policies:
        return self._try_object(Policies, 'Policies', '_policies')

    @property
    def api_tokens(self) -> APIToken:
        return self._try_object(APIToken, 'APIToken', '_api_tokens')

    @property
    def site_separation(self) -> SiteSeparation:
        return self._try_object(SiteSeparation, 'SiteSeparation', '_site_separation')

    @property
    def discovery(self) -> Discovery:
        return self._try_object(Discovery, 'Discovery', '_discovery')

    @property
    def authentication(self) -> Authentication:
        return self._try_object(Authentication, 'Authentication', '_authentication')

    @property
    def seeds(self) -> Seeds:
        return self._try_object(Seeds, 'Seeds', '_seeds')

    @property
    def vendor_api(self) -> VendorAPI:
        return self._try_object(VendorAPI, 'VendorAPI', '_vendor_api')

    @property
    def global_attributes(self) -> Attributes:
        if self._global_attributes is None:
            self._global_attributes = Attributes(client=self.client)
        return self._global_attributes



import json
from typing import Optional, List, Any, Dict, Union
from urllib.parse import quote_plus
from collections import defaultdict

from pydantic import BaseModel, PrivateAttr, TypeAdapter, computed_field, Field, FilePath

from ipfabric.tools import raise_for_status

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

OAS_DIR = files("ipfabric.oas")
CONTENT_TYPE = "application/json"

SEARCH_PATHS_IGNORE = [
    "/tables/networks/ipv6-routes",  # Route tables match too many items
    "/tables/networks/routes",
    "/tables/addressing/path-lookup-sources",
    "/tables/addressing/path-lookup-sources-multicast",
    "/tables/addressing/path-lookup-sources-unicast",
    "/tables/management/changes/managed-devs",
    "/tables/reports/discovery-tasks",
    "/tables/interfaces/connectivity-matrix/unmanaged-neighbors/summary",  # Summary table
    "/tables/routing/protocols/bgp/address-families",
    "/tables/routing/protocols/ospf/interfaces",
    "/tables/routing/protocols/ospf-v3/interfaces",
    "/tables/management/connectivity-errors",
]


class NestedColumn(BaseModel):
    parent: str
    child: str


class Scan(BaseModel):
    columns: Optional[List[Union[str, NestedColumn]]] = Field(default_factory=list)
    full_scan: bool = True


class Column(BaseModel):
    field: str
    filters: Optional[str] = None


class SubColumn(BaseModel):
    filters: Optional[str] = None
    array: Optional[bool] = Field(False)
    children: List[Column] = Field(default_factory=list)

    @property
    def children_by_name(self) -> Dict[str, Column]:
        return {_.field: _ for _ in self.children}

    @property
    def children_by_filters(self) -> Dict[str, List[Column]]:
        tmp = defaultdict(list)
        [tmp[_.filters].append(_) for _ in self.children]
        return dict(tmp)


class Endpoint(BaseModel):
    full_api_endpoint: str
    web_endpoint: Optional[str] = None
    web_url: Optional[str] = None
    columns: Optional[List[str]] = None
    nested_columns: Optional[Dict[str, SubColumn]] = Field(default_factory=dict)
    array_columns: Optional[Dict[str, SubColumn]] = Field(default_factory=dict)
    summary: Optional[str] = None
    title: Optional[str] = None
    tags: Optional[List[str]] = Field(default_factory=list)
    tag_groups: Optional[List[str]] = Field(default_factory=list)
    method: str
    ui_columns: Optional[List[str]] = None
    api_scope_id: Optional[str] = None
    description: Optional[str] = None
    ipv4: Optional[Scan] = Field(default_factory=Scan)
    ipv6: Optional[Scan] = Field(default_factory=Scan)
    mac: Optional[Scan] = Field(default_factory=Scan)

    def __repr__(self):
        return f"({self.method}, {self.api_endpoint})"

    @property
    def api_endpoint(self):
        return self.full_api_endpoint[1:]

    @property
    def web_menu(self):
        if self.tag_groups and self.tags:
            return f"{self.tag_groups[0]} — {self.tags[0]}"
        return None

    def filter_url(self, filters: Union[str, dict]) -> Union[None, str]:
        if not self.web_url:
            return None
        if isinstance(filters, str):
            filters = json.loads(filters)
        if "filters" not in filters:
            filters = {"filters": filters}
        return f"{self.web_url}?options=" + quote_plus(json.dumps(filters, separators=(",", ":")))


class Methods(BaseModel):
    full_api_endpoint: str
    get: Optional[Endpoint] = None
    put: Optional[Endpoint] = None
    patch: Optional[Endpoint] = None
    post: Optional[Endpoint] = None
    delete: Optional[Endpoint] = None

    @property
    def api_endpoint(self):
        return self.full_api_endpoint[1:]


class OAS(BaseModel):
    client: Any = Field(exclude=True)
    local_oas: bool = True
    local_oas_file: Optional[FilePath] = None
    _oas: Dict[str, Methods] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context) -> None:
        self._oas = self._get_oas()

    @property
    def oas(self) -> Dict[str, Methods]:
        return self._oas

    def _get_oas(self) -> Dict[str, Methods]:
        if not self.local_oas or (self.local_oas_file and self.local_oas):
            return self._parse_oas()
        try:
            min_oas = OAS_DIR.joinpath(self.client.api_version + ".json").read_text(encoding="UTF-8")
            oas = TypeAdapter(Dict[str, Methods]).validate_json(min_oas)
            return oas
        except FileNotFoundError:
            return self._parse_oas()

    @computed_field
    @property
    def web_to_api(self) -> Dict[str, Endpoint]:
        return {m.post.web_endpoint: m.post for m in self._oas.values() if m.post and m.post.web_endpoint}

    @computed_field
    @property
    def scope_to_api(self) -> Dict[str, Endpoint]:
        _ = dict()
        for methods in self._oas.values():
            for method in ["get", "put", "post", "patch", "delete"]:
                m = getattr(methods, method, None)
                if m and m.api_scope_id:
                    _[m.api_scope_id] = m
        return _

    def _complex_columns(self, data: Endpoint, spec: dict) -> Endpoint:
        try:
            x_columns = {_["key"]: _ for _ in spec["x-table"]["columns"]}
        except KeyError:
            x_columns = dict()
        try:
            columns = spec["responses"]["200"]["content"][CONTENT_TYPE]["schema"]["properties"]["data"]["items"][
                "properties"
            ]
            for k, v in columns.items():
                col_type = v.get("type", None)
                children = list()
                if k in x_columns and isinstance(x_columns[k].get("filter", None), dict):
                    children = [
                        Column(field=c["field"], filters=c["filter"]) for c in x_columns[k]["filter"]["children"]
                    ]
                if col_type == "array":
                    filters = v["filter"] if isinstance(v.get("filter", None), str) else None
                    data.array_columns[k] = SubColumn(array=True, children=children, filters=filters)
                    if v.get("items", {}).get("type", None) == "object":
                        data.nested_columns[k] = data.array_columns[k]
                elif col_type == "object":
                    data.nested_columns[k] = SubColumn(children=children)
        except KeyError:
            pass
        return self._global_search(data, spec)

    def _post_logic(self, data: Endpoint, spec: dict):
        try:
            data.web_endpoint = spec["x-table"]["webPath"]
            data.web_url = str(self.client.base_url.join(data.web_endpoint))
        except KeyError:
            pass
        try:
            data.title = spec["x-table"]["title"]
        except KeyError:
            pass
        try:
            data.ui_columns = [_["key"] for _ in spec["x-table"]["columns"]]
        except KeyError:
            pass
        try:
            columns = set(
                spec["requestBody"]["content"][CONTENT_TYPE]["schema"]["properties"]["columns"]["items"]["enum"]
            )
            data.columns = list(columns)
        except KeyError:
            pass
        return self._complex_columns(data, spec)

    def _complex_global_search(self, data: Endpoint) -> Endpoint:
        for cname, column in data.nested_columns.items():
            data.ipv4.columns.extend(
                [NestedColumn(parent=cname, child=_.field) for _ in column.children_by_filters.get("ip", [])]
            )
            data.ipv6.columns.extend(
                [NestedColumn(parent=cname, child=_.field) for _ in column.children_by_filters.get("ipv6", [])]
            )
            data.mac.columns.extend(
                [
                    NestedColumn(parent=cname, child=_.field)
                    for _ in column.children
                    if self._check_mac(_.field, data.full_api_endpoint)
                ]
            )
        return data

    @staticmethod
    def _check_mac(mac: str, endpoint: str) -> bool:
        return (
            "mac" in mac
            and mac not in ["macVerification", "macFlags", "macCount"]
            and endpoint not in ["/tables/inventory/devices"]
        )

    def _global_search(self, data: Endpoint, spec: dict) -> Endpoint:
        if not data.api_endpoint.startswith("tables") or data.full_api_endpoint in SEARCH_PATHS_IGNORE:
            return data
        try:
            columns = spec["x-table"]["columns"]
        except KeyError:
            return data
        for column in columns:
            if (
                column.get("filter", None) == "ip"
                and column["key"] not in ["loginIp"]
                and data.full_api_endpoint not in ["/tables/routing/protocols/ospf-v3/neighbors"]
            ):
                data.ipv4.columns.append(column["key"])
            elif column.get("filter", None) == "ipv6":
                data.ipv6.columns.append(column["key"])
            elif self._check_mac(column["key"], data.full_api_endpoint):
                data.mac.columns.append(column["key"])
        return self._complex_global_search(data)

    def _parse_oas(self) -> Dict[str, Methods]:
        if not self.local_oas or not self.local_oas_file:
            url = self.client.base_url.join(
                "/api/oas/openapi-extended.json"
                if self.client.os_api_version != "v6.9"
                else "/api/static/oas/openapi-fe.json"
            )  # TODO: 7.0 Fix for new OAS
            oas = raise_for_status(self.client.get(url, follow_redirects=True)).json()
        else:
            with open(self.local_oas_file, "r") as f:
                oas = json.load(f)

        endpoints = dict()
        for endpoint, methods in oas["paths"].items():
            methods_obj = Methods(full_api_endpoint=endpoint)
            for method, spec in methods.items():
                data = Endpoint(
                    full_api_endpoint=endpoint,
                    method=method,
                    api_scope_id=spec.get("x-apiScopeId", None),
                    summary=spec.get("summary", None),
                    description=spec.get("description", None),
                    tags=spec.get("tags", None),
                    tag_groups=spec.get("x-tagGroups", None),
                )
                if method == "post":
                    data = self._post_logic(data, spec)
                setattr(methods_obj, method, data)
            endpoints[endpoint[1:]] = methods_obj
        return endpoints



import logging
from typing import Any, Optional, List, Dict

from httpx import HTTPStatusError
from pydantic import Field, BaseModel

from ipfabric.models import User
from ipfabric.tools.shared import TIMEZONES, raise_for_status
from .rbac import Roles

logger = logging.getLogger("ipfabric")


class LocalUsers(BaseModel):
    client: Any = Field(exclude=True)
    _users: Optional[List[User]] = None
    _roles: Optional[Roles] = None

    def setup(self):
        self._users = self.get_users()
        try:
            self._roles = Roles(client=self.client)
        except HTTPStatusError:
            pass

    @property
    def users(self) -> List[User]:
        if not self._users:
            self.setup()
        return self._users

    @property
    def users_by_id(self) -> Dict[str, User]:
        return {_.user_id: _ for _ in self._users}

    def get_users(self, username: str = None) -> List[User]:
        """
        Gets all users or filters on one of the options.
        :param username: str: Username to filter
        :return: List of users
        """
        filters = {"username": ["ieq", username]} if username else None
        users = self.client.fetch_all("tables/users", filters=filters, snapshot=False)
        return [User(**user) for user in users]

    def get_user_by_id(self, user_id: str):
        """
        Gets a user by ID
        :param user_id: Union[str, int]: User ID to filter
        :return: User
        """
        resp = raise_for_status(self.client.get("users/" + str(user_id)))
        user = resp.json()
        return User(**user)

    def add_user(
        self,
        username: str,
        password: str,
        roles: list,
        timezone: str = "UTC",
    ):
        """
        Adds a user
        :param username: str: Username
        :param password: str: Must be 8 characters
        :param roles: list: Role IDs for Users
        :param timezone: str: v4.2 and above, Defaults UTC.  See pytz.all_timezones for correct syntax
        :return: User
        """
        if len(password) < 8:
            raise SyntaxError("Password must be 8 characters.")
        if self._roles and not all(x in [r.role_id for r in self._roles.roles] for x in roles):
            raise SyntaxError(f"Only accepted roles are {[r.role_id for r in self._roles.roles]}")
        payload = {
            "username": username,
            "password": password,
            "roleIds": roles,
        }
        if timezone.lower() not in TIMEZONES:
            raise ValueError(f"Timezone `{timezone}` is not located. Please see `ipfabric.tools.shared.TIMEZONES`.")
        payload["timezone"] = TIMEZONES[timezone.lower()]
        resp = raise_for_status(self.client.post("users", json=payload))
        self._users = self.get_users()
        return self.users_by_id[resp.json()["id"]]

    def delete_user(self, user_id: str) -> bool:
        """
        Deletes a user and returns list of remaining users
        :param user_id:
        :return:
        """
        raise_for_status(self.client.delete("users/" + str(user_id)))
        self._users = self.get_users()
        return True



import logging
from typing import Any, Optional, List

from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass

from ipfabric.exceptions import deprecated_args_decorator
from ipfabric.tools import raise_for_status

logger = logging.getLogger("ipfabric")


class Networks(BaseModel):
    exclude: List[str]
    include: List[str]


@deprecated_args_decorator(version="7.0", no_args=False, kwargs_only=True)
@dataclass
class Discovery:
    client: Any = Field(exclude=True)
    _networks: Optional[Networks] = None

    def __post_init__(self):
        self._networks = self._get_networks()

    @property
    def networks(self):
        return self._networks

    def _get_networks(self):
        res = raise_for_status(self.client.get("settings"))
        return Networks(**res.json()["networks"])

    def update_discovery_networks(self, subnets: list, include: bool = False):
        payload = dict()
        payload["networks"] = dict()
        if include:
            payload["networks"]["include"] = subnets
            payload["networks"]["exclude"] = self.networks.exclude
        else:
            payload["networks"]["exclude"] = subnets
            payload["networks"]["include"] = self.networks.include
        res = raise_for_status(self.client.patch("settings", json=payload))
        return Networks(**res.json()["networks"])



import json
import logging
import re
from typing import Any, Dict

from deepdiff import DeepDiff
from pydantic import BaseModel, Field, ConfigDict, PrivateAttr

from ipfabric.models import Policy, Role

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

logger = logging.getLogger("ipfabric")

EXCLUDE_PATH = re.compile(r"\['(policy_id|role_id)'\]")


class ManagedRBAC(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    ipf: Any = Field(exclude=True)
    _current_policies: dict = PrivateAttr(default_factory=dict)
    _policies: dict = PrivateAttr(default_factory=dict)
    _current_roles: dict = PrivateAttr(default_factory=dict)
    _roles: dict = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any) -> None:
        path = files("ipfabric.tools.managed_rbac.v" + str(self.ipf.os_version)[0:3])
        if not (path.joinpath("policies.json").is_file() and path.joinpath("roles.json").is_file()):
            logger.critical("No managed RBAC found for this version, please contact your SA.")
            exit()

        self._policies = json.loads(path.joinpath("policies.json").read_text())
        self._current_policies = {
            k: {"description": v.description, "paths": v.api_scope_ids, "policy_id": v.policy_id}
            for k, v in self.ipf.settings.policies.policies_by_name.items()
            if k in self._policies
        }
        self._roles = json.loads(path.joinpath("roles.json").read_text())
        self._current_roles = {
            k: {"description": v.description, "policies": [p.name for p in v.policies], "role_id": v.role_id}
            for k, v in self.ipf.settings.roles.roles_by_name.items()
            if k in self._roles
        }

    def _apply_policy(self) -> Dict[str, Policy]:
        policies = dict()
        if not self._current_policies:
            logger.warning("No Managed RBAC Policies found, pushing new policies.")
            policies = self.ipf.settings.policies.create_policies(
                [
                    {"name": k, "description": v["description"], "api_scope_ids": v["paths"]}
                    for k, v in self._policies.items()
                ]
            )
            return policies
        missing = {k: v for k, v in self._policies.items() if k not in self._current_policies}
        configured = {k: v for k, v in self._policies.items() if k in self._current_policies}
        if missing:
            logger.warning("Missing Managed RBAC Policies found, pushing new policies.")
            policies.update(
                self.ipf.settings.policies.create_policies(
                    [
                        {"name": k, "description": v["description"], "api_scope_ids": v["paths"]}
                        for k, v in missing.items()
                    ]
                )
            )
        if configured:
            diff = DeepDiff(configured, self._current_policies, exclude_regex_paths=EXCLUDE_PATH, ignore_order=True)
            if diff:
                logger.warning("Policy changes found, pushing updates.")
                for _ in diff.affected_root_keys.items:
                    logger.warning(f"Updating policy: {_}.")
                    policy = self.ipf.settings.policies.update_policy(
                        policy=self._current_policies[_]["policy_id"],
                        api_scope_ids=self._policies[_]["paths"],
                        name=_,
                        description=self._policies[_]["description"],
                        replace_api_scopes=True,
                    )
                    policies.update({_: policy})
                policies.update(
                    {
                        _: self.ipf.settings.policies.policies_by_name[_]
                        for _ in configured
                        if _ not in diff.affected_root_keys.items
                    }
                )
            else:
                policies.update({_: self.ipf.settings.policies.policies_by_name[_] for _ in configured})
        return policies

    def _apply_role(self, policies) -> Dict[str, Role]:
        roles = dict()
        if not self._roles:
            logger.warning("No Managed RBAC Roles found, pushing new roles.")
            roles = self.ipf.settings.roles.create_roles(
                [
                    {"name": k, "description": v["description"], "policy_ids": [policies[_] for _ in v["policies"]]}
                    for k, v in self._roles.items()
                ]
            )
            return roles
        missing = {k: v for k, v in self._roles.items() if k not in self._current_roles}
        configured = {k: v for k, v in self._roles.items() if k in self._current_roles}
        if missing:
            logger.warning("Missing Managed RBAC Roles found, pushing new roles.")
            roles.update(
                self.ipf.settings.roles.create_roles(
                    [
                        {"name": k, "description": v["description"], "policy_ids": [policies[_] for _ in v["policies"]]}
                        for k, v in missing.items()
                    ]
                )
            )
        if configured:
            diff = DeepDiff(configured, self._current_roles, exclude_regex_paths=EXCLUDE_PATH, ignore_order=True)
            if diff:
                logger.warning("Role changes found, pushing updates.")
                for _ in diff.affected_root_keys.items:
                    logger.warning(f"Updating role: {_}.")
                    role = self.ipf.settings.roles.update_role(
                        role=self._current_roles[_]["role_id"],
                        policy_ids=[policies[p] for p in self._roles[_]["policies"]],
                        name=_,
                        description=self._roles[_]["description"],
                        replace=True,
                    )
                    roles.update({_: role})
                roles.update(
                    {
                        _: self.ipf.settings.roles.roles_by_name[_]
                        for _ in configured
                        if _ not in diff.affected_root_keys.items
                    }
                )
            else:
                roles.update({_: self.ipf.settings.roles.roles_by_name[_] for _ in configured})
        return roles

    def apply(self):
        policies = self._apply_policy()
        self.ipf.settings.roles.update()
        roles = self._apply_role(policies)
        return policies, roles



import json
import logging
import re
from json import loads, dumps
from typing import Optional, Any, Union, Literal, List, Dict, overload

from httpx import InvalidURL, URL
from pydantic import ConfigDict
from pydantic.dataclasses import dataclass

from ipfabric.api import IPFabricAPI
from ipfabric.diagrams import Diagram
from ipfabric.models import Technology, Inventory, Jobs, Intent, Devices
from ipfabric.settings import Settings
from ipfabric.tools import TIMEZONES, raise_for_status
from .exceptions import deprecated_args_decorator

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

logger = logging.getLogger("ipfabric")

RE_PATH = re.compile(r"^/?(api/)?v\d(\.\d)?/")
RE_TABLE = re.compile(r"^tables/")
EXPORT_FORMAT = Literal["json", "csv", "df"]


@deprecated_args_decorator(version="7.0", no_args=False, kwargs_only=True)
@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class IPFClient(IPFabricAPI):
    _inventory: Optional[Inventory] = None
    _technology: Optional[Technology] = None
    _jobs: Optional[Jobs] = None
    _diagram: Optional[Diagram] = None
    _intent: Optional[Intent] = None
    _devices: Optional[Devices] = None
    _settings: Optional[Settings] = None

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "timeout":
            self._client.timeout = value
        elif name == "snapshot_id":
            value = self._switch_snapshot(value)
            self.__dict__[name] = value
            if value and self.snapshot.disabled_intent_verification is False and self._intent:
                self._intent = Intent(client=self)
            if value and self._devices:
                self._devices = self.load_devices()
            return None
        super().__setattr__(name, value)

    def __repr__(self):
        return f"IPFClient(base_url='{str(self.base_url)}')"

    @property
    def inventory(self) -> Inventory:
        self._inventory = self._inventory if self._inventory else Inventory(client=self)
        return self._inventory

    @property
    def technology(self) -> Technology:
        self._technology = self._technology if self._technology else Technology(client=self)
        return self._technology

    @property
    def jobs(self) -> Jobs:
        self._jobs = self._jobs if self._jobs else Jobs(client=self)
        return self._jobs

    @property
    def diagram(self) -> Diagram:
        self._diagram = self._diagram if self._diagram else Diagram(ipf=self)
        return self._diagram

    @property
    def intent(self) -> Intent:
        self._intent = self._intent if self._intent else Intent(client=self)
        return self._intent

    @property
    def settings(self) -> Settings:
        self._settings = self._settings if self._settings else Settings(client=self)
        return self._settings

    @property
    def devices(self) -> Devices:
        """get devices"""
        if not self._devices:
            logger.info("Devices not loaded, loading devices.")
            self._devices = self.load_devices()
        return self._devices

    @devices.setter
    def devices(self, devices: Devices):
        self._devices = devices

    def update(self) -> None:
        self._snapshots.update()
        if self.snapshot_id not in self.loaded_snapshots:
            logger.warning(f"Snapshot {self.snapshot_id} is no longer loaded switching to `$last`.")
            self.snapshot_id = "$last"

    def load_devices(self, device_filters: dict = None, device_attr_filters: dict = None) -> Devices:
        devices = Devices(
            client=self,
            snapshot_id=self.snapshot_id,
            device_filters=device_filters,
            device_attr_filters=device_attr_filters,
        )
        return devices

    def _check_url(self, url) -> str:
        path = URL(url).path
        path = path if path[0] == "/" else "/" + path
        if path in self.web_to_api:
            return self.web_to_api[path].api_endpoint
        r = RE_PATH.search(path)
        url = path[r.end():] if r else path  # fmt: skip
        url = url[1:] if url[0] == "/" else url
        return url

    def _create_payload(self, url, snapshot_id, filters, sort, attr_filters) -> dict:
        payload = {"format": {"dataType": "json"}}
        if filters and isinstance(filters, str):
            payload["filters"] = loads(filters)
        elif filters and isinstance(filters, dict):
            payload["filters"] = filters
        if snapshot_id:
            payload["snapshot"] = snapshot_id
        if sort:
            payload["sort"] = sort
        if RE_TABLE.match(url) and (attr_filters or self.attribute_filters):
            payload["attributeFilters"] = attr_filters or self.attribute_filters
        return payload

    def _check_url_payload(self, url, snapshot_id, filters, reports, sort, attr_filters, export, csv_tz):
        url = self._check_url(url)
        oas_data = getattr(self.oas.get(url, {}), "post", None)
        payload = self._create_payload(url, snapshot_id, filters, sort, attr_filters)

        if export != "csv":
            if isinstance(reports, (str, list)):
                payload["reports"] = reports
            elif reports is True and oas_data and oas_data.web_endpoint:
                payload["reports"] = oas_data.web_endpoint
            elif reports is True and (not oas_data or not oas_data.web_endpoint):
                logger.warning(
                    f"Could not automatically discover Web Endpoint for Intent Data for table '/{url}'.\n"
                    f"Returning results without Intent Rules."
                )
        else:
            if reports:
                logger.warning("CSV export does not return reports, parameter has been excluded.")
            payload["format"] = {"exportToFile": True, "dataType": "csv"}
            if csv_tz and csv_tz.lower() not in TIMEZONES:
                raise ValueError(
                    f"CSV timezone '{csv_tz}' not in available timezones; see `ipfabric.tools.shared.TIMEZONES`"
                )
            elif csv_tz:
                payload["format"]["options"] = {"timezone": TIMEZONES[csv_tz.lower()]}
        return url, payload

    def _get_payload(self, url, payload) -> Union[str, bool]:
        snapshot_id = payload.pop("snapshot", None)
        reports = payload.pop("reports") if "reports" in payload and isinstance(payload["reports"], str) else None
        p = "&".join([f"{k}={dumps(v, separators=(',', ':'))}" for k, v in payload.items()])
        if snapshot_id:
            p += f"&snapshot={snapshot_id}"
        if reports:
            p += f"&reports={reports}"
        url = self.base_url.join(url + f"?{p}")
        if len(str(url)) > 4096:
            return False
        return url

    def _stream(self, url, payload, export):
        get_url = self._get_payload(url, payload)
        if get_url is False and export == "csv":
            raise InvalidURL("URL exceeds max character limit of 4096 cannot export to CSV.")
        elif get_url is False:
            logger.warning("URL exceeds max character limit of 4096 switching to pagination.")
            return False
        with self.stream("GET", get_url) as stream_resp:
            stream_resp.raise_for_status()
            data = stream_resp.read()
        return data if export == "csv" else loads(data)["data"]

    def query(self, url: str, payload: Union[str, dict], get_all: bool = True) -> Union[List[dict], bytes]:
        """Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.

        Args:
            url: Example: https://demo1.ipfabric.io/api/v1/tables/vlan/device-summary or tables/vlan/device-summary
            payload: Dictionary to submit in POST or can be JSON string (i.e. read from file).
            get_all: Default use pager to get all results and ignore pagination information in the payload

        Returns:
            list or bytes: List of Dictionary objects or bytes if csv.
        """
        url = self._check_url(url)
        if isinstance(payload, str):
            payload = loads(payload)
        export = payload.get("format", {}).get("dataType", "json")
        data = False
        if export == "csv" or (self.streaming and get_all):
            data = self._stream(url, payload, export)
        elif get_all and data is False:
            data = self._ipf_pager(url, payload)
        elif data is False:
            res = raise_for_status(self.post(url, json=payload))
            data = res.json()["data"]
        return data

    def get_columns(self, url: str, ui: bool = False) -> List[str]:
        """Checks OAS to find available columns.

        Args:
            url: API url to post
            ui: True to return columns available in the UI, default False

        Returns:
            list: List of column names
        """
        ctype = "ui_columns" if ui else "columns"
        oas_data = getattr(self.oas.get(url, {}), "post", None)
        if oas_data and getattr(oas_data, ctype):
            return getattr(oas_data, ctype)
        else:
            raise LookupError(f"Could not find {'UI' if ui else 'API'} columns for table '/{url}'.")

    def get_count(
        self,
        url: str,
        filters: Optional[Union[dict, str]] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        snapshot: bool = True,
    ) -> int:
        """Get a total number of rows
        Args:
            url: API URL to post to
            filters: Optional dictionary of filters
            attr_filters: Optional dictionary of attribute filters
            snapshot_id: Optional snapshot_id to override default
            snapshot: Set to False for some tables like management endpoints.
        Returns:
            int: a count of rows
        """
        snapshot_id = snapshot_id or self.snapshot_id if snapshot else None
        url, payload = self._check_url_payload(url, snapshot_id, filters, None, None, attr_filters, None, None)
        payload.update({"columns": ["id"], "pagination": {"limit": 1, "start": 0}})
        res = raise_for_status(self.post(url, json=payload))
        return res.json()["_meta"]["count"]

    def _fetch_setup(self, url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot):
        if export == "df" and DataFrame is None:
            raise ImportError("pandas not installed. Run `pip install ipfabric[pd]`.")
        snapshot_id = snapshot_id or self.snapshot_id if snapshot else None
        if "color" in dumps(filters) and not reports:
            reports = True
        url, payload = self._check_url_payload(url, snapshot_id, filters, reports, sort, attr_filters, export, csv_tz)
        payload["columns"] = columns or self.get_columns(url)
        return url, payload

    @overload
    def fetch(
        self,
        url,
        export: Literal["json"] = ...,
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> List[dict]: ...

    @overload
    def fetch(
        self,
        url,
        export: Literal["csv"],
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        url,
        export: Literal["df"] = ...,
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> DataFrame: ...

    def fetch(
        self,
        url,
        export: EXPORT_FORMAT = "json",
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ):
        """Gets data from IP Fabric for specified endpoint

        Args:
            url: Example tables/vlan/device-summary
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional list of columns to return, None will return all
            filters: Optional dictionary of filters
            limit: Default to 1,000 rows
            start: Starts at 0
            snapshot_id: Optional snapshot_id to override default
            reports: String of frontend URL where the reports are displayed or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            attr_filters: Optional dictionary to apply an Attribute filter
            snapshot: Set to False for some tables like management endpoints.
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pandas.DataFrame]: List of dict if json, bytes string if CSV, DataFrame is df
        """
        url, payload = self._fetch_setup(
            url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot
        )
        payload["pagination"] = dict(start=start, limit=limit)
        data = False
        if export == "csv" or self.streaming:
            data = self._stream(url, payload, export)
        if data is False:
            res = raise_for_status(self.post(url, json=payload))
            data = res.json()["data"]
        if export == "df":
            data = DataFrame.from_records(data) if export == "df" else data
        return data

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["json"] = ...,
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> List[dict]: ...

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["csv"],
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["df"],
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
    ) -> DataFrame: ...

    def fetch_all(
        self,
        url: str,
        export: EXPORT_FORMAT = "json",
        columns: Optional[List] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from IP Fabric for specified endpoint

        Args:
            url: Example tables/vlan/device-summary
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional list of columns to return, None will return all
            filters: Optional dictionary of filters
            snapshot_id: Optional snapshot_id to override default
            reports: String of frontend URL where the reports are displayed or a list of report IDs
            sort: Optional dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            attr_filters: Optional dictionary to apply an Attribute filter
            snapshot: Set to False for some tables like management endpoints.
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pandas.DataFrame]: List of dict if json, bytes string if CSV, DataFrame is df
        """
        url, payload = self._fetch_setup(
            url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot
        )
        data = False
        if export == "csv" or self.streaming:
            data = self._stream(url, payload, export)
        if data is False:
            data = self._ipf_pager(url, payload)
        if export == "df":
            data = DataFrame.from_records(data)
        return data

    def _shared_url(self, url: Union[int, str], table: bool = True) -> dict:
        try:
            url_id = str(int(url))
        except ValueError:
            parsed_url = URL(url)
            if table and parsed_url.params.get("copyId"):
                url_id = parsed_url.params.get("copyId")
            elif not table:
                url_id = parsed_url.path.split("/")[-1]
            else:
                raise SyntaxError("Wrong endpoint selected.")
        if table:
            resp = self.get(f"/tables/url/{url_id}")
        else:
            resp = self.get(f"/graphs/urls/{url_id}")
        return raise_for_status(resp).json()

    def shared_view(self, url: Union[int, str], data=True, reports: bool = True) -> Union[str, list]:
        """Takes a shared table view link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or full/partial URL (`/inventory/devices?copyId=1453653298`)
            data: Defaults to return the data instead of printing the code
            reports: True to return Intent Verification data

        Returns: List of dictionaries of the data or a string representing the code.
        """
        query = self._shared_url(url, True)
        url = self.web_to_api[query["webEndpoint"]].api_endpoint
        hidden = {k for k, v in query.get("columnVisibility", {}).items() if not v}
        columns = {k for k, v in query.get("columnVisibility", {}).items() if v}
        columns.update({_ for _ in query.get("columnWidth", {}) if _ not in hidden})
        columns.update({_ for _ in self.get_columns(url, ui=True) if _ not in hidden})
        columns = sorted(list(columns))

        filters, sort = query.get("filters", {}), query.get("sort", {})
        filters.pop("selectedFilter", None)

        if data:
            return self.fetch_all(
                url=url, columns=columns, filters=filters, sort=sort, reports=query["webEndpoint"] if reports else False
            )
        code = f"{url.split('/')[-1]} = ipf.fetch_all(\n    {url=},\n    {columns=},"
        code += f"\n    {filters=}," if filters else ""
        code += f"\n    {sort=}," if sort else ""
        code += f"\n    reports='{query['webEndpoint']}',\n)" if reports else "\n)"
        return code

    def share_link(
        self,
        url: str,
        filters: Union[dict, str],
        columns: Optional[List] = None,
        sort: Optional[dict] = None,
    ) -> str:
        """Creates a shareable link for a table.

        Args:
            url: Example tables/vlan/device-summary
            filters: Filter dictionary to create view for
            columns: Optional list of columns to display, defaults to all. Some columns are hidden in the UI and will
                     not be displayed.
            sort: Dict of column to sort on
        Returns:
            URL string
        """
        url = self._check_url(url)
        try:
            web_endpoint = self.oas[url].post.web_endpoint
        except KeyError:
            raise NotImplementedError(f"Cannot create Shared View URL for table `{url}` as it has no web endpoint.")
        _columns = {_: True for _ in self.get_columns(url)}

        if columns:
            _columns.update({_: False for _ in _columns if _ not in columns})
        try:
            hidden = list(
                {k for k, v in _columns.items() if v and k != "id"}.difference(set(self.get_columns(url, ui=True)))
            )
            if hidden:
                logger.warning(
                    f"Column(s) `{hidden}` are hidden in the UI and will not be displayed using the shared view link."
                )
        except LookupError:
            pass
        try:
            user = self.get("tables/settings", params={"name": web_endpoint}).json()["userSettings"]
            auto_size, col_width = user.get("autoSizedColumns", True), user.get("columnWidth", {})
        except KeyError:
            auto_size, col_width = True, {}
        payload = {
            "autoSizedColumns": auto_size,
            "columnVisibility": _columns,
            "webEndpoint": web_endpoint,
            "filters": loads(filters) if isinstance(filters, str) else filters,
        }
        payload.update({"columnWidth": col_width} if col_width else {})
        payload.update({"sort": sort} if sort else {})
        resp = raise_for_status(self.post("tables/url", json=payload))
        return str(self.base_url.join(web_endpoint)) + f"?copyId={resp.json()['id']}"

    def post_to_get(self, url: str, payload: Union[dict, str], ignore_reports: bool = False):
        """

        Args:
            url: str: API or Frontend URL
            payload: Union[str, dict]: Copy/Paste payload from API description in UI, only required object is 'snapshot'
            ignore_reports: bool: Remove the reports from the payload (intent data) from being included.

        Returns: str: Get URL to return table data instead of pagination

        """
        if isinstance(payload, str):
            payload = json.loads(payload)
        snapshot = payload.get("snapshot", None)
        url, payload = self._fetch_setup(
            url,
            "json",
            payload.get("columns", None),
            snapshot,
            payload.get("filters", None),
            None if ignore_reports else payload.get("reports", None),
            payload.get("sort", None),
            payload.get("attributeFilters", None),
            None,
            True if snapshot else False,
        )
        return str(self._get_payload(url, payload))



import logging
from typing import Union, Dict, List, Optional, Literal, Any

from pydantic import BaseModel

from ipfabric.tools import raise_for_status
from .input_models import (
    Unicast,
    Multicast,
    Host2GW,
    Network,
    NetworkSettings,
    PathLookupSettings,
    Overlay,
    SharedView,
    GraphSettings,
    GroupSettings,
)
from .output_models.graph_result import NetworkEdge, Node, PathLookupEdge, GraphResult, PathLookup, Position

logger = logging.getLogger("ipfabric")
GRAPHS_URL = "graphs/"


class Diagram(BaseModel):
    ipf: Any

    def _check_snapshot_id(self, snapshot_id):
        snapshot_id = snapshot_id or self.ipf.snapshot_id
        if snapshot_id not in self.ipf.loaded_snapshots:
            raise ValueError(f"Snapshot {snapshot_id} is not loaded or not found in IP Fabric.")
        return snapshot_id

    def _intent_overlay(self, overlay: Overlay) -> dict:
        try:
            Overlay._valid_intentrule(overlay["intentRuleId"])
        except ValueError:
            if not self.ipf.intent.loaded:
                self.ipf.intent.load_intent()
            intents = self.ipf.intent.intents_by_name[overlay.intentRuleId]
            if len(intents) > 1:
                raise ValueError(f"Multiple Intents found with name `{overlay['intentRuleId']}`.")
            else:
                overlay.intentRuleId = intents[0].intent_id
        return overlay.overlay()

    def _snapshot_overlay(self, overlay: Overlay, snapshot_id: str = None) -> dict:
        if overlay.snapshotToCompare not in self.ipf.loaded_snapshots:
            raise ValueError(f"Snapshot `{overlay.snapshotToCompare}` is not loaded.")
        overlay.snapshotToCompare = self.ipf.snapshots[overlay.snapshotToCompare].snapshot_id
        if snapshot_id or self.ipf.snapshot_id == overlay.snapshotToCompare:
            raise ValueError(f"Cannot compare snapshot `{overlay.snapshotToCompare}` to itself.")
        self._check_snapshot_id(overlay.snapshotToCompare)
        return overlay.overlay()

    def _format_overlay(self, overlay: Union[Overlay, dict], snapshot_id: str = None) -> dict:
        if isinstance(overlay, dict):
            overlay = Overlay(**overlay)
        if overlay.intentRuleId:
            return self._intent_overlay(overlay)
        return self._snapshot_overlay(overlay, snapshot_id)

    def _query(
        self,
        parameters: dict,
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        image: Literal["png", "svg", "json"] = "json",
        graph_settings: dict = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        positions: Optional[Dict[str, Union[Position, dict]]] = None,
    ):
        """
        Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.
        :param parameters: dict: Dictionary to submit in POST.
        :return: list: List of Dictionary objects.
        """
        url = GRAPHS_URL
        if image in ["svg", "png"]:
            url += image
        payload = dict(parameters=parameters, snapshot=self._check_snapshot_id(snapshot_id))
        if overlay:
            payload["overlay"] = self._format_overlay(overlay, snapshot_id)
        if graph_settings:
            payload["settings"] = graph_settings
        if attr_filters or self.ipf.attribute_filters:
            payload["attributeFilters"] = attr_filters or self.ipf.attribute_filters
        if positions:
            payload["positions"] = {k: dict(v) for k, v in positions.items()}
        res = raise_for_status(self.ipf.post(url, json=payload))
        return res.json() if image == "json" else res.content

    def json(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
        positions: Optional[Dict[str, Union[Position, dict]]] = None,
    ) -> dict:
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            image="json",
            overlay=overlay,
            attr_filters=attr_filters,
            graph_settings=graph_settings.settings() if graph_settings else None,
            positions=positions,
        )

    def svg(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
        positions: Optional[Dict[str, Union[Position, dict]]] = None,
    ) -> bytes:
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="svg",
            graph_settings=graph_settings.settings() if graph_settings else None,
            positions=positions,
        )

    def png(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
        positions: Optional[Dict[str, Union[Position, dict]]] = None,
    ) -> bytes:
        return self._query(
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters(),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="png",
            graph_settings=graph_settings.settings() if graph_settings else None,
            positions=positions,
        )

    def share_link(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Union[Overlay, dict] = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
        positions: Optional[Dict[str, Union[Position, dict]]] = None,
    ) -> str:
        parameters = (
            parameters.parameters(unicast_swap_src_dst) if isinstance(parameters, Unicast) else parameters.parameters()
        )

        resp = self._query(
            parameters,
            snapshot_id=snapshot_id,
            overlay=overlay,
            image="json",
            attr_filters=attr_filters,
            graph_settings=graph_settings.settings() if graph_settings else None,
            positions=positions,
        )

        parameters.pop("layouts", None)
        payload = {
            "graphView": {
                "name": "Shared view",
                "parameters": parameters,
                "collapsedNodeGroups": [],
                "hiddenNodes": [],
                "positions": {k: v["position"] for k, v in resp["graphResult"]["graphData"]["nodes"].items()},
                "settings": resp["graphResult"]["settings"],
            },
            "snapshot": self._check_snapshot_id(snapshot_id),
        }
        if overlay:
            payload["graphView"]["overlay"] = self._format_overlay(overlay, snapshot_id)
        res = raise_for_status(self.ipf.post("graphs/urls", json=payload))
        return str(self.ipf.base_url.join(f"/diagrams/share/{res.json()['id']}"))

    def model(
        self,
        parameters: Union[Unicast, Multicast, Host2GW, Network],
        snapshot_id: str = None,
        overlay: Overlay = None,
        graph_settings: Union[NetworkSettings, PathLookupSettings] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        unicast_swap_src_dst: bool = False,
        positions: Optional[Dict[str, Union[Position, dict]]] = None,
    ) -> GraphResult:
        json_data = self.json(
            parameters, snapshot_id, overlay, graph_settings, attr_filters, unicast_swap_src_dst, positions
        )
        edge_setting_dict = self._diagram_edge_settings(json_data["graphResult"]["settings"])
        if isinstance(parameters, Network):
            return self._diagram_network(json_data, edge_setting_dict)
        else:
            return self._diagram_pathlookup(json_data, edge_setting_dict)

    @staticmethod
    def _diagram_network(json_data: dict, edge_setting_dict: dict, pathlookup: bool = False) -> GraphResult:
        edges, nodes = dict(), dict()
        for node_id, node in json_data["graphResult"]["graphData"]["nodes"].items():
            nodes[node_id] = Node.model_validate(node)

        for edge_id, edge_json in json_data["graphResult"]["graphData"]["edges"].items():
            edge = PathLookupEdge.model_validate(edge_json) if pathlookup else NetworkEdge.model_validate(edge_json)
            edge.protocol = (
                edge_setting_dict[edge.edgeSettingsId].name if edge.edgeSettingsId in edge_setting_dict else None
            )
            if edge.source:
                edge.source = nodes[edge.source]
            if edge.target:
                edge.target = nodes[edge.target]
            edges[edge_id] = edge

        return GraphResult(edges=edges, nodes=nodes)

    def _diagram_pathlookup(self, json_data: dict, edge_setting_dict: dict) -> GraphResult:
        graph_result = self._diagram_network(json_data, edge_setting_dict, pathlookup=True)

        for edge_id, edge in graph_result.edges.items():
            for prev_id in edge.prevEdgeIds:
                edge.prevEdge.append(graph_result.edges[prev_id])
            for next_id in edge.nextEdgeIds:
                edge.nextEdge.append(graph_result.edges[next_id] if next_id in graph_result.edges else next_id)
        graph_result.pathlookup = PathLookup.model_validate(json_data["pathlookup"])

        return graph_result

    @staticmethod
    def _diagram_edge_settings(graph_settings: dict) -> dict:
        net_settings = GraphSettings(**graph_settings)
        edge_setting_dict = dict()
        for edge in net_settings.edges:
            edge_setting_dict[edge.id] = edge
            if isinstance(edge, GroupSettings):
                for child in edge.children:
                    edge_setting_dict[child.id] = child
        return edge_setting_dict

    def shared_view(
        self,
        url: Union[int, str],
        image: Literal["json", "code", "model", "svg", "png"] = "json",
        positions: bool = False,
    ):
        """Takes a shared graph link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or full/partial URL (`/diagrams/share/1453626097`)
            image: Defaults to return the data instead of printing the code
            positions: If returning code then include positions of nodes in example; Default False

        Returns: The graph data or string representing the code to produce it.
        """
        query = self.ipf._shared_url(url, False)
        if query["snapshot"] not in self.ipf.loaded_snapshots:
            logger.warning(f'Snapshot {query["snapshot"]} is not loaded, switching to {self.ipf.snapshot_id}.')
            query["snapshot"] = self.ipf.snapshot_id
        overlay = query["graphView"].get("overlay", {})
        if (
            overlay
            and overlay.get("snapshotToCompare", None)
            and overlay["snapshotToCompare"] not in self.ipf.loaded_snapshots
        ):
            logger.error(f"Snapshot `{overlay['snapshotToCompare']}` is not loaded to compare, removing overlay.")
            overlay = {}

        view = SharedView(
            snapshot_id=query["snapshot"],
            client_snapshot_id=self.ipf.snapshot_id,
            hidden_nodes=query["graphView"]["hiddenNodes"],
            collapsed_node_groups=query["graphView"]["collapsedNodeGroups"],
            positions=query["graphView"]["positions"],
            settings=query["graphView"]["settings"]["edges"],
            params=query["graphView"]["parameters"],
            hidden_devs=query["graphView"]["settings"].get("hiddenDeviceTypes", []),
            path_lookup=query["graphView"]["settings"].get("pathLookup", None),
            overlay=overlay,
        )
        if view.hidden_nodes or view.collapsed_node_groups:
            logger.warning("Hidden Nodes and Collapsed Node Groups are only available via the UI.")

        if image in ["json", "svg", "png"]:
            return self._query(
                parameters=query["graphView"]["parameters"],
                snapshot_id=query["snapshot"],
                overlay=overlay,
                image=image,
                graph_settings=query["graphView"]["settings"],
                positions=query["graphView"]["positions"],
            )

        if image == "code":
            return "\n".join(view.create_code(positions))
        if image == "model":
            return self.model(
                parameters=view.graph_model(),
                snapshot_id=query["snapshot"],
                overlay=overlay,
                graph_settings=view.graph_settings(),
                positions=query["graphView"]["positions"],
            )
        return None



import logging
from typing import Any, Optional
from warnings import warn

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class Sdwan(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    @computed_field
    @property
    def sites(self) -> Table:
        # TODO: Remove 7.0
        warn("Sdwan.sites will be moved to Sdwan.versa_sites in 7.0", DeprecationWarning, stacklevel=2)
        return Table(client=self.client, endpoint="tables/sdwan/sites", sn=self.sn)

    @computed_field
    @property
    def links(self) -> Table:
        # TODO: Remove 7.0
        warn("Sdwan.links will be moved to Sdwan.versa_links in 7.0", DeprecationWarning, stacklevel=2)
        return Table(client=self.client, endpoint="tables/sdwan/links", sn=self.sn)

    @computed_field
    @property
    def versa_sites(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/links", sn=self.sn)

    @computed_field
    @property
    def versa_links(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/links", sn=self.sn)



import logging
from collections import defaultdict
from typing import Any, Union, List, Dict, Optional
from warnings import warn

from httpx import HTTPStatusError
from pydantic import BaseModel, PrivateAttr, Field

from ipfabric.tools import raise_for_status
from .intent_check import Group, IntentCheck

logger = logging.getLogger("ipfabric")
COLOR_DICT = dict(nan=-1, green=0, blue=10, amber=20, red=30)


class Intent(BaseModel):
    client: Any = Field(exclude=True)
    _intent_checks: List[IntentCheck] = PrivateAttr(default_factory=list)
    _groups: List[Group] = PrivateAttr(default_factory=list)
    _snapshot_id: Optional[str] = None

    def model_post_init(self, __context: Any) -> None:
        self._snapshot_id = self.client.snapshot_id
        if self.snapshot_id:
            self._intent_checks: List[IntentCheck] = self.get_intent_checks()
            self._groups: List[Group] = self.get_groups()

    @property
    def intent_checks(self):
        return self._intent_checks

    @property
    def groups(self):
        return self._groups

    @property
    def snapshot_id(self):
        return self._snapshot_id

    @snapshot_id.setter
    def snapshot_id(self, snapshot_id: str):
        self._snapshot_id = snapshot_id

    def get_intent_checks(self, snapshot_id: str = None) -> List[IntentCheck]:
        """Gets all intent checks and returns a list of them.

        Args:
        snapshot_id: Optional snapshot ID to get different data

        Returns:
            list: List of intent checks
        """
        snapshot = self.client.snapshots[snapshot_id] if snapshot_id else self.client.snapshots[self.snapshot_id]
        if not snapshot.loaded:
            raise ValueError(f"Snapshot {snapshot.snapshot_id} is not loaded; cannot pull Intent Rules.")
        if snapshot.disabled_intent_verification is True:
            raise ValueError(
                f"Snapshot {snapshot.snapshot_id} has Intent Verification computation disabled; "
                "cannot pull Intent Rules."
            )
        res = raise_for_status(self.client.get("reports", params=dict(snapshot=snapshot.snapshot_id)))
        try:
            return [IntentCheck(**check) for check in res.json()]
        except HTTPStatusError:
            logger.warning(self.client._api_insuf_rights + 'on GET "/reports". Will not load Intents.')
            return list()

    def load_intent(self, snapshot_id: str = None):
        """Loads intent checks into the class.

        Args:
            snapshot_id: Uses a different Snapshot ID then client
        """
        self.snapshot_id = snapshot_id or self.client.snapshot_id
        self._intent_checks = self.get_intent_checks(snapshot_id)
        self._groups = self.get_groups()

    def get_groups(self) -> list:
        """
        Returns:
            list: list of groups
        """
        res = self.client.get("reports/groups")
        if res.status_code == 200:
            return [Group(**group) for group in res.json()]
        else:  # TODO: Fix this error
            logger.warning(self.client._api_insuf_rights + 'on GET "/reports/groups". Will not load Intent Groups.')
            return list()

    @property
    def custom(self) -> List[IntentCheck]:
        return [c for c in self.intent_checks if c.custom]

    @property
    def builtin(self) -> List[IntentCheck]:
        return [c for c in self.intent_checks if not c.custom]

    @property
    def intent_by_id(self) -> Dict[str, IntentCheck]:
        return {c.intent_id: c for c in self.intent_checks}

    @property
    def intents_by_name(self) -> Dict[str, List[IntentCheck]]:
        reports = defaultdict(list)
        [reports[c.name].append(c) for c in self.intent_checks]
        return dict(reports)

    @property
    def intent_by_name(self) -> Dict[str, IntentCheck]:  # TODO: remove 7.0
        msg = (
            "This method will be removed in v7.0. "
            "Please use IPFClient().intent.intents_by_name() as Intent Names are not unique."
        )
        warn(
            msg,
            DeprecationWarning,
            stacklevel=2,
        )
        logger.warning(msg)
        return {c.name: c for c in self.intent_checks}

    @property
    def intent_ids_by_web_endpoint(self) -> Dict[str, List[str]]:
        reports = defaultdict(list)
        [reports[c.web_endpoint].append(c.intent_id) for c in self.intent_checks]
        return dict(reports)

    @property
    def intent_ids_by_api_endpoint(self) -> Dict[str, List[str]]:
        reports = defaultdict(list)
        [reports[c.api_endpoint].append(c.intent_id) for c in self.intent_checks]
        return dict(reports)

    @property
    def group_by_id(self) -> Dict[str, Group]:
        return {g.group_id: g for g in self.groups}

    @property
    def group_by_name(self) -> Dict[str, Group]:
        return {g.name: g for g in self.groups}

    def get_results(self, intent: IntentCheck, color: Union[str, int], snapshot_id: str = None) -> list:
        """Get the outcome of an Intent Check by a specific color

        Args:
            intent: an IntentCheck, please see the Intent Check Model
            color: color of intent check
            snapshot_id: Uses a different Snapshot ID then client

        Returns:
            list: List of Dictionary objects.
        """
        if isinstance(color, str):
            color = COLOR_DICT[color]
        return self._get_data(intent, snapshot_id or self.snapshot_id, color)

    def get_all_results(self, intent: IntentCheck, snapshot_id: str = None):
        """set the intent check attributes

        Args:
            intent: an IntentCheck, please see the Intent Check Model
            snapshot_id: Uses a different Snapshot ID then client

        Returns:
            list: List of Dictionary objects.
        """
        snapshot_id = snapshot_id or self.snapshot_id
        for color_str, color_int in COLOR_DICT.items():
            if color_int == -1 or getattr(intent.result.checks, color_str):
                setattr(intent.result_data, color_str, self._get_data(intent, snapshot_id, color_int))
        return intent

    def _get_data(self, intent: IntentCheck, snapshot_id: str, color: int):
        return self.client.fetch_all(
            intent.api_endpoint,
            snapshot_id=snapshot_id,
            reports=intent.web_endpoint,
            filters={intent.column: ["color", "eq", color]},
        )

    def compare_snapshot(self, snapshot_id: str, reverse: bool = False) -> list:
        """Compares all intents against another snapshot.
        Current is the snapshot loaded into the class
        Other is the snapshot specified in this method.  Use reverse=True to flip them.

        Args:
            snapshot_id: Snapshot ID to compare against this will be the "other" key
            reverse: Default False, setting to true will flip current and other.
        Returns:
            list: List of dictionaries
        """
        new_intents = {i.name: i for i in self.get_intent_checks(snapshot_id)}
        comparison = list()
        for name, intent in new_intents.items():
            old = self.intents_by_name[name][0].result
            compare = intent.result.compare(old) if reverse else old.compare(intent.result)
            for desc, value in compare.items():
                n = desc if desc != "count" else "total"
                comparison.append({"name": name, "id": intent.intent_id, "check": n, **value})
        return comparison



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class ManagedNetworks(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    @computed_field
    @property
    def networks(self) -> Table:
        return Table(client=self.client, endpoint="tables/networks", sn=self.sn)

    @computed_field
    @property
    def gateway_redundancy(self) -> Table:
        return Table(client=self.client, endpoint="tables/networks/gateway-redundancy", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class Neighbors(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    @computed_field
    @property
    def neighbors_all(self) -> Table:
        return Table(client=self.client, endpoint="tables/neighbors/all", sn=self.sn)

    @computed_field
    @property
    def neighbors_unmanaged(self) -> Table:
        return Table(client=self.client, endpoint="tables/neighbors/unmanaged", sn=self.sn)

    @computed_field
    @property
    def neighbors_unidirectional(self) -> Table:
        return Table(client=self.client, endpoint="tables/neighbors/unidirectional", sn=self.sn)

    @computed_field
    @property
    def neighbors_endpoints(self) -> Table:
        return Table(client=self.client, endpoint="tables/neighbors/endpoints", sn=self.sn)



import logging
from typing import Any, Optional
from warnings import warn

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class LoadBalancing(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    @computed_field
    @property
    def virtual_servers(self) -> Table:
        return Table(client=self.client, endpoint="tables/load-balancing/virtual-servers", sn=self.sn)

    @computed_field
    @property
    def virtual_servers_pools(self) -> Table:
        return Table(client=self.client, endpoint="tables/load-balancing/virtual-servers/pools", sn=self.sn)

    @computed_field
    @property
    def virtual_servers_pool_members(self) -> Table:
        return Table(client=self.client, endpoint="tables/load-balancing/virtual-servers/pool-members", sn=self.sn)

    @computed_field
    @property
    def partitions(self) -> Table:
        return Table(client=self.client, endpoint="tables/load-balancing/partitions", sn=self.sn)

    @computed_field
    @property
    def virtual_servers_f5_partitions(self) -> Table:
        # TODO: Remove 7.0
        warn(
            "LoadBalancing.virtual_servers_f5_partitions will be moved to LoadBalancing.partitions in 7.0",
            DeprecationWarning,
            stacklevel=2,
        )
        return Table(client=self.client, endpoint="tables/load-balancing/f5-partitions", sn=self.sn)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from warnings import warn

if TYPE_CHECKING:
    from ipfabric.models.users import User

logger = logging.getLogger("ipfabric")


def api_insuf_rights(user: User):
    msg = f'API_INSUFFICIENT_RIGHTS for user "{user.username}" '
    if user.token:
        msg += f'token "{user.token.description}" '
    return msg


def deprecated_args_decorator(  # noqa: C901
    version: str, arg_type=None, no_args: bool = True, kwargs_only: bool = False
):

    def inner(func):

        fname = func.__name__

        def wrapper(*args, **kwargs):
            if arg_type:
                arg_types = [arg_type] if isinstance(arg_type, str) else arg_type
                msg = (
                    f"`{fname}()` no longer accepts parameter of `{arg_type}` and will be removed in version {version}."
                )
                if args:
                    for arg in args:
                        if type(arg).__name__ in arg_types:
                            warn(msg, DeprecationWarning, stacklevel=2)
                            logger.warning(msg)
                if kwargs:
                    for arg in kwargs.values():
                        if type(arg).__name__ in arg_types:
                            warn(msg, DeprecationWarning, stacklevel=2)
                            logger.warning(msg)
            elif no_args and (len(args) > 1 or kwargs):
                warn(
                    f"`{fname}()` parameters are deprecated and will be removed in version {version}.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                logger.warning(f"`{fname}()` parameters are deprecated and will be removed in version {version}.")
            elif kwargs_only and len(args):
                msg = (
                    f"`{fname}()` positional arguments are deprecated and will be removed in version {version}. "
                    f"Please switch to keyword arguments only."
                )
                warn(msg, DeprecationWarning, stacklevel=2)
                logger.warning(msg)
            return func(*args, **kwargs)

        return wrapper

    return inner



import logging
from typing import Any, Optional, List, Dict, Union

from httpx import HTTPStatusError
from pydantic import BaseModel, Field, PrivateAttr

from ipfabric.models.oas import Endpoint
from ipfabric.models.rbac import Role, Policy
from ipfabric.tools import valid_slug, raise_for_status

logger = logging.getLogger("ipfabric")


class Policies(BaseModel):
    client: Any = Field(exclude=True)
    _policies: Optional[List[Policy]] = PrivateAttr(None)

    def model_post_init(self, __context: Any) -> None:
        self.update()

    def update(self):
        self._policies = self.get_policies()

    @property
    def policies(self):
        return self._policies

    @property
    def policies_by_id(self) -> Dict[str, Policy]:
        return {_.policy_id: _ for _ in self.policies}

    @property
    def policies_by_name(self) -> Dict[str, Policy]:
        return {_.name: _ for _ in self.policies}

    def get_policies(self, policy_name: str = None) -> Union[List[Policy], Policy, None]:
        """
        Gets all policies or filters on one of the options.
        :param policy_name: str: Case Sensitive Policy Name to filter
        :return: List of policies or single policy if name passed
        """
        filters = {"name": ["eq", policy_name]} if policy_name else None
        policies = list()
        for policy in self.client.fetch_all("tables/policies", filters=filters, snapshot=False):
            p = Policy(**policy)
            p.api_scopes = [self.client.scope_to_api[_] for _ in p.api_scope_ids if _ in self.client.scope_to_api]
            policies.append(p)
        if policy_name:
            return policies.pop() if policies else None
        return policies

    def search_policies_for_endpoint(self, api_scope_id: str) -> List[Policy]:
        return [p for p in self.policies if api_scope_id in p.api_scopes_by_id]

    def _check_policy(self, name=None, description=None, api_scope_ids=None, attribute_filters=None):
        if not (name and description):
            raise SyntaxError("No Policy name or description provided.")
        elif name in self.policies_by_name:
            raise ValueError(f"Duplicate Policy name detected: `{name}`.")
        elif api_scope_ids and attribute_filters:
            raise SyntaxError("Policy must either be API Scoped or Attribute Filter not both.")
        elif not (api_scope_ids or attribute_filters):
            raise SyntaxError("No API Scopes or Attribute Filters provided.")
        if api_scope_ids:
            return self._fmt_api_scopes(api_scope_ids)
        return None

    @staticmethod
    def _fmt_api_scopes(api_scope_ids):
        tmp = set()
        for _ in api_scope_ids:
            if isinstance(_, Endpoint):
                tmp.add(_.api_scope_id)
            elif isinstance(_, str):
                tmp.add(str(_))
            else:
                raise ValueError(f"Invalid Scope: {_}")
        return list(tmp)

    def _create_policy(
        self,
        name: str,
        description: str,
        api_scope_ids: List[Union[int, str, Endpoint]] = None,
        attribute_filters: dict = None,
    ) -> Policy:
        fmt_api_scopes = self._check_policy(name, description, api_scope_ids, attribute_filters)
        payload = {
            "name": valid_slug(name),
            "description": description,
        }
        if api_scope_ids:
            payload["apiScopeIds"] = fmt_api_scopes
        else:
            payload["attributeFilters"] = attribute_filters

        raise_for_status(self.client.post(f"policies/scopes/{'api' if api_scope_ids else 'attributes'}", json=payload))
        return self.get_policies(name)

    def create_policy(
        self,
        name: str,
        description: str,
        api_scope_ids: List[Union[int, str, Endpoint]] = None,
        attribute_filters: dict = None,
    ) -> Policy:
        policy = self._create_policy(name, description, api_scope_ids, attribute_filters)
        self.update()
        return policy

    def create_policies(
        self, policies: List[Dict[str, Union[str, dict, List[Union[int, str, Endpoint]]]]]
    ) -> Dict[str, Policy]:
        """Create multiple policies.

        Args:
            policies: List of dicts [{name: str, description: str, api_scope_ids: List[Union[int, str, Endpoint]]}
                                    {name: str, description: str, attribute_filters: dict}]

        Returns: Dict of created policies by name.
        """
        created, policy_names = dict(), set()
        for policy in policies:
            self._check_policy(**policy)
            policy_names.add(policy["name"])
        if len(policy_names) != len(policies):
            raise ValueError("Duplicate policy names found.")

        for policy in policies:
            try:
                created[policy["name"]] = self._create_policy(**policy)
            except HTTPStatusError:
                created[policy["name"]] = False
        self.update()
        return created

    def _delete_policy(self, policy: Union[int, str, Policy]) -> bool:
        """Deletes a Role.

        Args:
            policy: Policy ID or Policy object

        Returns: True
        """
        raise_for_status(
            self.client.delete(f"policies/{policy.policy_id if isinstance(policy, Policy) else str(policy)}")
        )
        return True

    def delete_policy(self, policy: Union[int, str, Policy]) -> bool:
        _ = self._delete_policy(policy)
        self.update()
        return _

    def delete_policies(self, policies: List[Union[int, str, Policy]]) -> Dict[str, bool]:
        deleted = dict()
        for policy in policies:
            policy_id = policy.policy_id if isinstance(policy, Policy) else str(policy)
            try:
                deleted[policy_id] = self._delete_policy(policy_id)
            except HTTPStatusError:
                deleted[policy_id] = False
        self.update()
        return deleted

    def update_policy(  # noqa: C901
        self,
        policy: Union[int, str, Policy],
        api_scope_ids: List[Union[int, str, Endpoint]] = None,
        attribute_filters: dict = None,
        name: str = None,
        description: str = None,
        replace_api_scopes: bool = False,
    ) -> Policy:
        if not (api_scope_ids or attribute_filters or name or description):
            raise SyntaxError("No updates provided.")
        policy = policy if isinstance(policy, Policy) else self.policies_by_id[str(policy)]
        if policy.scope_type == "attributeScopes" and api_scope_ids:
            raise SyntaxError(f"Policy `{policy.name}` is Attribute Scoped not API Scoped.")
        elif policy.scope_type == "apiScopes" and attribute_filters:
            raise SyntaxError(f"Policy `{policy.name}` is API Scoped not Attribute Scoped.")

        payload = {
            "description": description or policy.description,
            "name": name or policy.name,
        }
        if api_scope_ids:
            api_scope_ids = set(self._fmt_api_scopes(api_scope_ids))
            if not replace_api_scopes:
                api_scope_ids.update(set(policy.api_scope_ids))
            payload["apiScopeIds"] = list(api_scope_ids)
        else:
            payload["attributeFilters"] = attribute_filters

        resp = raise_for_status(
            self.client.patch(
                f"policies/scopes/{'api' if policy.scope_type == 'apiScopes' else 'attributes'}/{policy.policy_id}",
                json=payload,
            )
        )
        self.update()
        return self.policies_by_id[resp.json()["id"]]


class Roles(BaseModel):
    client: Any
    _roles: Optional[List[Role]] = None
    _policies: Optional[Policies] = PrivateAttr(None)

    def model_post_init(self, __context: Any) -> None:
        self._policies = self.client.settings.policies
        self._roles = self.get_roles()

    def update(self):
        self._policies.update()
        self._roles = self.get_roles()

    @property
    def roles(self):
        return self._roles

    @property
    def roles_by_id(self):
        return {r.role_id: r for r in self.roles}

    @property
    def roles_by_name(self):
        return {r.name: r for r in self.roles}

    def get_roles(self, role_name: str = None) -> [List[Role], Role, None]:
        """
        Gets all roles or filters on one of the options.
        :param role_name: str: Case Sensitive Role Name to filter
        :return: List of roles or single role if filtered
        """
        filters = {"name": ["ieq", role_name]} if role_name else None
        roles = list()
        for role in self.client.fetch_all("tables/roles", filters=filters, snapshot=False):
            roles.append(
                Role(**role, policies=[self._policies.policies_by_id[_] for _ in role["policyIds"]])
                if self._policies
                else Role(**role)
            )
        if role_name:
            return roles.pop() if roles else None
        return roles

    def search_roles_for_policy(self, policy_name: str = None, policy_id: Union[str, int] = None) -> List[Role]:
        if policy_name and not self._policies:
            raise ValueError(self.client._api_insuf_rights + 'on POST "/tables/policies".')
        if policy_name:
            return [r for r in self.roles if policy_name in r.policies_by_name]
        elif policy_id:
            return [r for r in self.roles if str(policy_id) in r.policies_by_id]
        else:
            raise SyntaxError("No Policy Name or ID provided.")

    def search_roles_for_endpoint(self, api_scope_id: str) -> List[Role]:
        policies = self._policies.search_policies_for_endpoint(api_scope_id)
        if not policies:
            logger.warning(f"No policies found for API Scope ID {api_scope_id}.")
            return list()
        roles = set()
        for p in policies:
            roles.update(set(self.search_roles_for_policy(policy_id=p.policy_id)))
        return list(roles)

    def _create_role(self, name: str, description: str, policy_ids: List[Union[int, str, Policy]]) -> Role:
        payload = {
            "name": valid_slug(name),
            "description": description,
            "policyIds": [_.policy_id if isinstance(_, Policy) else str(_) for _ in policy_ids],
        }
        raise_for_status(self.client.post("roles", json=payload))
        return self.get_roles(name)

    def create_role(self, name: str, description: str, policy_ids: List[Union[int, str, Policy]]) -> Role:
        role = self._create_role(name, description, policy_ids)
        self.update()
        return role

    def create_roles(self, roles: List[Dict[str, Union[str, List[Union[int, str, Policy]]]]]) -> Dict[str, Role]:
        """Create multiple roles.

        Args:
            roles: List of dicts [{name: str, description: str, policy_ids: List[Union[int, str, Policy]]}]

        Returns: Dict of created roles by name.
        """
        created, role_names = dict(), set()
        for role in roles:
            if set(role.keys()) != {"name", "description", "policy_ids"} or not (
                isinstance(role["name"], str)
                and isinstance(role["description"], str)
                and isinstance(role["policy_ids"], list)
            ):
                raise SyntaxError(f"Invalid role `{str(role)}`.")
            elif role["name"] in self.roles_by_name:
                raise ValueError(f"Duplicate Role name detected: `{role['name']}`.")
            role_names.add(role["name"])
        if len(role_names) != len(roles):
            raise ValueError("Duplicate role names found.")

        for role in roles:
            try:
                created[role["name"]] = self.create_role(**role)
            except HTTPStatusError:
                created[role["name"]] = False
        return created

    def _delete_role(self, role: Union[int, str, Role]) -> bool:
        """Deletes a Role.

        Args:
            role: Role ID or Role object

        Returns: True
        """
        raise_for_status(self.client.delete(f"roles/{role.role_id if isinstance(role, Role) else str(role)}"))
        return True

    def delete_role(self, role: Union[int, str, Role]) -> bool:
        _ = self._delete_role(role)
        self.update()
        return _

    def delete_roles(self, roles: List[Union[int, str, Role]]) -> Dict[str, bool]:
        deleted = dict()
        for role in roles:
            role_id = role.role_id if isinstance(role, Role) else str(role)
            try:
                deleted[role_id] = self._delete_role(role_id)
            except HTTPStatusError:
                deleted[role_id] = False
        return deleted

    def update_role(
        self,
        role: Union[int, str, Role],
        policy_ids: List[Union[int, str, Policy]] = None,
        name: str = None,
        description: str = None,
        replace: bool = False,
    ) -> Role:
        """Updates a Role.

        Args:
            role: Role ID or Role object
            policy_ids: List of Policy IDs or Policy Objects
            name: New name for role, if None then do not update
            description: New description, if None then do not update
            replace: True to replace all policies, defaults to false to append policies

        Returns: Role object
        """
        if not (policy_ids or name or description):
            raise SyntaxError("No updates provided.")
        role = role if isinstance(role, Role) else self.roles_by_id[str(role)]
        policies = set(role.policy_ids)
        if replace and policy_ids:
            policies = {_.policy_id if isinstance(_, Policy) else str(_) for _ in policy_ids}
        elif policy_ids:
            policies.update({_.policy_id if isinstance(_, Policy) else str(_) for _ in policy_ids})
        payload = {
            "description": description or role.description,
            "name": name or role.name,
            "policyIds": list(policies),
        }
        resp = raise_for_status(self.client.put(f"roles/{role.role_id}", json=payload))
        self.update()
        return self.roles_by_id[resp.json()["id"]]



import logging
from copy import deepcopy
from typing import Any, Optional
from warnings import warn

from pydantic import Field
from pydantic.dataclasses import dataclass

from ipfabric.exceptions import deprecated_args_decorator
from ipfabric.tools import raise_for_status

logger = logging.getLogger("ipfabric")


@deprecated_args_decorator(version="7.0", no_args=False, kwargs_only=True)
@dataclass
class SiteSeparation:
    client: Optional[Any] = Field(None, exclude=True)
    ipf: Optional[Any] = None

    def __post_init__(self):
        if not self.ipf and not self.client:
            raise SyntaxError("No IPF Client passed.")
        elif self.ipf and not self.client:
            msg = "Argument `ipf` is being changed to `client` in `7.0` please update your scripts."
            warn(msg, DeprecationWarning, stacklevel=2)
            logger.warning(msg)
            self.client = self.ipf

    def get_separation_rules(self):
        return self.client.get("settings/site-separation").json()

    def _post_rule(self, data):
        return raise_for_status(self.client.post("settings/site-separation/test-regex", json=data)).json()

    @staticmethod
    def _create_rule(transformation, regex):
        transformation = transformation.lower()
        if transformation not in ["uppercase", "lowercase", "none"]:
            raise SyntaxError('Transformation type is not in ["uppercase", "lowercase", "none"].')
        return {"regex": regex, "transformation": transformation}

    def get_rule_matches(self, rule):
        rule = deepcopy(rule)
        [rule.pop(key, None) for key in ["id", "note", "siteName"]]
        return self._post_rule(rule)

    def get_hostname_matches(self, regex, transformation):
        rule = self._create_rule(transformation, regex)
        rule["type"] = "regexHostname"
        return self._post_rule(rule)

    def get_snmp_matches(self, regex, transformation):
        rule = self._create_rule(transformation, regex)
        rule["type"] = "regexSnmpLocation"
        return self._post_rule(rule)
